from functools import cached_property
from pathlib import Path

from databricks.labs.blueprint.installation import Installation
from databricks.labs.lsql.backends import RuntimeBackend, SqlBackend
from databricks.sdk import WorkspaceClient, core

from databricks.labs.ucx.__about__ import __version__
from databricks.labs.ucx.assessment.clusters import ClustersCrawler, PoliciesCrawler
from databricks.labs.ucx.assessment.init_scripts import GlobalInitScriptCrawler
from databricks.labs.ucx.assessment.jobs import JobsCrawler, SubmitRunsCrawler
from databricks.labs.ucx.assessment.pipelines import PipelinesCrawler
from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.ucx.contexts.application import GlobalContext
from databricks.labs.ucx.hive_metastore import TablesInMounts
from databricks.labs.ucx.hive_metastore.table_size import TableSizeCrawler
from databricks.labs.ucx.hive_metastore.tables import FasterTableScanCrawler
from databricks.labs.ucx.installer.logs import TaskRunWarningRecorder
from databricks.labs.ucx.progress.workflow_runs import WorkflowRunRecorder


class RuntimeContext(GlobalContext):
    @cached_property
    def _config_path(self) -> Path:
        config = self.named_parameters.get("config")
        if not config:
            raise ValueError("config flag is required")
        return Path(config)

    @cached_property
    def config(self) -> WorkspaceConfig:
        return Installation.load_local(WorkspaceConfig, self._config_path)

    @cached_property
    def connect_config(self) -> core.Config:
        connect = self.config.connect
        assert connect, "connect is required"
        return connect

    @cached_property
    def workspace_client(self) -> WorkspaceClient:
        return WorkspaceClient(config=self.connect_config, product='ucx', product_version=__version__)

    @cached_property
    def sql_backend(self) -> SqlBackend:
        return RuntimeBackend(debug_truncate_bytes=self.connect_config.debug_truncate_bytes)

    @cached_property
    def installation(self) -> Installation:
        install_folder = self._config_path.parent.as_posix().removeprefix("/Workspace")
        return Installation(self.workspace_client, "ucx", install_folder=install_folder)

    @cached_property
    def jobs_crawler(self) -> JobsCrawler:
        return JobsCrawler(self.workspace_client, self.sql_backend, self.inventory_database)

    @cached_property
    def submit_runs_crawler(self) -> SubmitRunsCrawler:
        return SubmitRunsCrawler(
            self.workspace_client,
            self.sql_backend,
            self.inventory_database,
            self.config.num_days_submit_runs_history,
        )

    @cached_property
    def clusters_crawler(self) -> ClustersCrawler:
        return ClustersCrawler(self.workspace_client, self.sql_backend, self.inventory_database)

    @cached_property
    def pipelines_crawler(self) -> PipelinesCrawler:
        return PipelinesCrawler(self.workspace_client, self.sql_backend, self.inventory_database)

    @cached_property
    def table_size_crawler(self) -> TableSizeCrawler:
        return TableSizeCrawler(self.sql_backend, self.inventory_database, self.config.include_databases)

    @cached_property
    def policies_crawler(self) -> PoliciesCrawler:
        return PoliciesCrawler(self.workspace_client, self.sql_backend, self.inventory_database)

    @cached_property
    def global_init_scripts_crawler(self) -> GlobalInitScriptCrawler:
        return GlobalInitScriptCrawler(self.workspace_client, self.sql_backend, self.inventory_database)

    @cached_property
    def tables_crawler(self):
        # TODO: Update tables crawler inheritance to specify return type hint
        return FasterTableScanCrawler(self.sql_backend, self.inventory_database)

    @cached_property
    def tables_in_mounts(self) -> TablesInMounts:
        return TablesInMounts(
            self.sql_backend,
            self.workspace_client,
            self.inventory_database,
            self.mounts_crawler,
            self.config.include_mounts,
            self.config.exclude_paths_in_mount,
            self.config.include_paths_in_mount,
        )

    @cached_property
    def task_run_warning_recorder(self) -> TaskRunWarningRecorder:
        return TaskRunWarningRecorder(
            self._config_path.parent,
            self.named_parameters["workflow"],
            int(self.named_parameters["job_id"]),
            int(self.named_parameters["parent_run_id"]),
            self.sql_backend,
            self.inventory_database,
            int(self.named_parameters.get("attempt", "0")),
        )

    @cached_property
    def workflow_run_recorder(self) -> WorkflowRunRecorder:
        return WorkflowRunRecorder(
            self.sql_backend,
            self.config.ucx_catalog,
            workspace_id=self.workspace_client.get_workspace_id(),
            workflow_name=self.named_parameters["workflow"],
            workflow_id=int(self.named_parameters["job_id"]),
            workflow_run_id=int(self.named_parameters["parent_run_id"]),
            workflow_run_attempt=int(self.named_parameters.get("attempt", 0)),
            workflow_start_time=self.named_parameters["start_time"],
        )



import logging
import subprocess

logger = logging.getLogger(__name__)


def escape_sql_identifier(path: str, *, maxsplit: int = 2) -> str:
    """
    Escapes the path components to make them SQL safe.

    Args:
        path (str): The dot-separated path of a catalog object.
        maxsplit (int): The maximum number of splits to perform.

    Returns:
         str: The path with all parts escaped in backticks.
    """
    if not path:
        return path
    parts = path.split(".", maxsplit=maxsplit)
    escaped = [f"`{part.strip('`').replace('`', '``')}`" for part in parts]
    return ".".join(escaped)


def run_command(command: str | list[str]) -> tuple[int, str, str]:
    args = command.split() if isinstance(command, str) else command
    logger.info(f"Invoking command: {args!r}")
    with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as process:
        output, error = process.communicate()
        return process.returncode, output.decode("utf-8"), error.decode("utf-8")



from collections.abc import Iterator

from databricks.labs.lsql.backends import SqlBackend
from databricks.labs.lsql.core import Row

from .base import (
    DataComparator,
    DataComparisonResult,
    TableIdentifier,
    DataProfiler,
    DataProfilingResult,
)


class StandardDataComparator(DataComparator):
    _DATA_COMPARISON_QUERY_TEMPLATE = """
    WITH compare_results AS (
        SELECT
            CASE
                WHEN source.hash_value IS NULL AND target.hash_value IS NULL THEN TRUE
                WHEN source.hash_value IS NULL OR target.hash_value IS NULL THEN FALSE
                WHEN source.hash_value = target.hash_value THEN TRUE
                ELSE FALSE
            END AS is_match,
            CASE
                WHEN target.hash_value IS NULL THEN 1
                ELSE 0
            END AS target_missing_count,
            CASE
                WHEN source.hash_value IS NULL THEN 1
                ELSE 0
            END AS source_missing_count
        FROM (
            SELECT {source_hash_expr} AS hash_value
            FROM {source_table_fqn}
        ) AS source
        FULL OUTER JOIN (
            SELECT {target_hash_expr} AS hash_value
            FROM {target_table_fqn}
        ) AS target
        ON source.hash_value = target.hash_value
    )
    SELECT
        COUNT(*) AS total_mismatches,
        COALESCE(SUM(target_missing_count), 0) AS target_missing_count,
        COALESCE(SUM(source_missing_count), 0) AS source_missing_count
    FROM compare_results
    WHERE is_match IS FALSE;
    """

    def __init__(self, sql_backend: SqlBackend, data_profiler: DataProfiler):
        self._sql_backend = sql_backend
        self._data_profiler = data_profiler

    def compare_data(
        self,
        source: TableIdentifier,
        target: TableIdentifier,
        compare_rows: bool = False,
    ) -> DataComparisonResult:
        """
        This method compares the data of two tables. It takes two TableIdentifier objects as input, which represent
        the source and target tables for which the data are to be compared.

        Note: This method does not handle exceptions raised during the execution of the SQL query or
        the retrieval of the table metadata. These exceptions are expected to be handled by the caller in a manner
        appropriate for their context.
        """
        source_data_profile = self._data_profiler.profile_data(source)
        target_data_profile = self._data_profiler.profile_data(target)
        if not compare_rows:
            return DataComparisonResult(
                source_row_count=source_data_profile.row_count,
                target_row_count=target_data_profile.row_count,
            )
        comparison_query = self._build_data_comparison_query(
            source_data_profile,
            target_data_profile,
        )
        query_result: Iterator[Row] = self._sql_backend.fetch(comparison_query)
        count_row = next(query_result)
        target_missing_count = int(count_row["target_missing_count"])
        source_missing_count = int(count_row["source_missing_count"])
        return DataComparisonResult(
            source_row_count=source_data_profile.row_count,
            target_row_count=target_data_profile.row_count,
            target_missing_count=target_missing_count,
            source_missing_count=source_missing_count,
        )

    @classmethod
    def _build_data_comparison_query(
        cls,
        source_data_profile: DataProfilingResult,
        target_data_profile: DataProfilingResult,
    ) -> str:
        source_table = source_data_profile.table_metadata.identifier
        target_table = target_data_profile.table_metadata.identifier
        source_hash_inputs = _build_data_comparison_hash_inputs(source_data_profile)
        target_hash_inputs = _build_data_comparison_hash_inputs(target_data_profile)
        comparison_query = cls._DATA_COMPARISON_QUERY_TEMPLATE.format(
            source_hash_expr=f"SHA2(CONCAT_WS('|', {', '.join(source_hash_inputs)}), 256)",
            target_hash_expr=f"SHA2(CONCAT_WS('|', {', '.join(target_hash_inputs)}), 256)",
            source_table_fqn=source_table.fqn_escaped,
            target_table_fqn=target_table.fqn_escaped,
        )

        return comparison_query


def _build_data_comparison_hash_inputs(data_profile: DataProfilingResult) -> list[str]:
    source_metadata = data_profile.table_metadata
    inputs = []
    for column in source_metadata.columns:
        data_type = column.data_type.lower()
        transformed_column = column.name

        if data_type.startswith("array"):
            transformed_column = f"TO_JSON(SORT_ARRAY({column.name}))"
        elif data_type.startswith("map") or data_type.startswith("struct"):
            transformed_column = f"TO_JSON({column.name})"

        inputs.append(f"COALESCE(TRIM({transformed_column}), '')")
    return inputs



import datetime as dt
import logging
from collections.abc import Iterable
from concurrent.futures import FIRST_COMPLETED, ThreadPoolExecutor, wait
from itertools import groupby

from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import InternalError, NotFound, ResourceDoesNotExist
from databricks.sdk.retries import retried
from databricks.sdk.service.workspace import ObjectInfo, ObjectType

logger = logging.getLogger(__name__)


class WorkspaceListing:
    def __init__(
        self,
        ws: WorkspaceClient,
        num_threads: int,
        verify_timeout: dt.timedelta | None = dt.timedelta(minutes=1),
        *,
        with_directories: bool = True,
    ):
        self.start_time = None
        self._ws = ws
        self.results: list[ObjectInfo] = []
        self._num_threads = num_threads
        self._with_directories = with_directories
        self._counter = 0
        self._verify_timeout = verify_timeout

    def _progress_report(self, _):
        self._counter += 1
        measuring_time = dt.datetime.now()
        delta_from_start = measuring_time - self.start_time
        rps = self._counter / delta_from_start.total_seconds()
        directory_count = len([r for r in self.results if r.object_type == ObjectType.DIRECTORY])
        other_count = len([r for r in self.results if r.object_type != ObjectType.DIRECTORY])
        if self._counter % 10 == 0:
            logger.info(
                f"Made {self._counter} workspace listing calls, "
                f"collected {len(self.results)} objects ({directory_count} dirs and {other_count} other objects),"
                f" rps: {rps:.3f}/sec"
            )

    def _list_workspace(self, path: str) -> Iterable[ObjectInfo]:
        list_retry_on_value_error = retried(on=[InternalError], timeout=self._verify_timeout)
        list_retried_check = list_retry_on_value_error(self._ws.workspace.list)
        return list_retried_check(path=path, recursive=False)

    def _list_and_analyze(self, obj: ObjectInfo) -> tuple[list[ObjectInfo], list[ObjectInfo]]:
        directories = []
        others = []
        if not obj.path:
            return [], []
        try:
            list_workspace = self._list_workspace(obj.path)

            def is_dir(obj: ObjectInfo) -> bool:
                return obj.object_type == ObjectType.DIRECTORY

            grouped_iterator: Iterable[tuple[bool, Iterable[ObjectInfo]]] = groupby(list_workspace, key=is_dir)
            for is_directory, objects in grouped_iterator:
                if is_directory:
                    directories.extend(list(objects))
                else:
                    others.extend(list(objects))
            logger.debug(f"Listed {obj.path}, found {len(directories)} sub-directories and {len(others)} other objects")
        except NotFound:
            # See https://github.com/databrickslabs/ucx/issues/230
            logger.warning(f"{obj.path} is not listable. Ignoring")
        except TimeoutError as e:
            logger.warning(f"Could not list {obj.path} due to backend error: {e}")
        return directories, others

    def walk(self, start_path="/"):
        self.start_time = dt.datetime.now()
        logger.info(f"Recursive WorkspaceFS listing started at {self.start_time}")
        try:
            root_object = self._ws.workspace.get_status(start_path)
        except ResourceDoesNotExist:
            logger.warning(f"removed on the backend {start_path}")
            return self.results
        self.results.append(root_object)
        with ThreadPoolExecutor(self._num_threads) as executor:
            initial_future = executor.submit(self._list_and_analyze, root_object)
            initial_future.add_done_callback(self._progress_report)
            futures_to_objects = {initial_future: root_object}
            while futures_to_objects:
                futures_done, _ = wait(futures_to_objects, return_when=FIRST_COMPLETED)
                for future in futures_done:
                    futures_to_objects.pop(future)
                    directories, others = future.result()
                    self.results.extend(directories)
                    self.results.extend(others)
                    if not directories:
                        continue
                    new_futures = {}
                    for directory in directories:
                        new_future = executor.submit(self._list_and_analyze, directory)
                        new_future.add_done_callback(self._progress_report)
                        new_futures[new_future] = directory
                    futures_to_objects.update(new_futures)
            logger.info(
                f"Recursive WorkspaceFS listing finished at {dt.datetime.now()}. "
                f"Total time taken for workspace listing: {dt.datetime.now() - self.start_time}"
            )
            self._progress_report(None)
        return self.results



import logging
from dataclasses import dataclass, field

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import BadRequest
from databricks.sdk.errors.platform import InvalidParameterValue, PermissionDenied, NotFound
from databricks.sdk.service.catalog import (
    AzureManagedIdentityRequest,
    AzureServicePrincipal,
    Privilege,
    StorageCredentialInfo,
    ValidationResultResult,
)

from databricks.labs.ucx.assessment.azure import AzureServicePrincipalCrawler
from databricks.labs.ucx.assessment.secrets import SecretsMixin
from databricks.labs.ucx.azure.access import (
    AzureResourcePermissions,
    StoragePermissionMapping,
)

logger = logging.getLogger(__name__)


# A dataclass to host service_principal info and its client_secret info
@dataclass
class ServicePrincipalMigrationInfo:
    permission_mapping: StoragePermissionMapping
    client_secret: str


@dataclass
class StorageCredentialValidationResult:
    name: str
    application_id: str
    read_only: bool | None
    validated_on: str
    directory_id: str | None = None  # str when storage credential created for AzureServicePrincipal
    failures: list[str] = field(default_factory=list)

    @classmethod
    def _get_application_and_directory_id(
        cls, storage_credential_info: StorageCredentialInfo
    ) -> tuple[str, str | None]:
        if storage_credential_info.azure_service_principal is not None:
            application_id = storage_credential_info.azure_service_principal.application_id
            directory_id = storage_credential_info.azure_service_principal.directory_id
            return application_id, directory_id

        if storage_credential_info.azure_managed_identity is not None:
            if storage_credential_info.azure_managed_identity.managed_identity_id is not None:
                return storage_credential_info.azure_managed_identity.managed_identity_id, None
            return storage_credential_info.azure_managed_identity.access_connector_id, None

        raise KeyError("Storage credential info is missing an application id.")

    @classmethod
    def from_storage_credential_info(
        cls,
        storage_credential_info: StorageCredentialInfo,
        validated_on: str,
        failures: list[str],
    ):
        assert storage_credential_info.name is not None
        application_id, directory_id = cls._get_application_and_directory_id(storage_credential_info)
        return cls(
            storage_credential_info.name,
            application_id,
            storage_credential_info.read_only,
            validated_on,
            directory_id,
            failures,
        )


class StorageCredentialManager:
    def __init__(self, ws: WorkspaceClient):
        self._ws = ws

    def list(self, include_names: set[str] | None = None) -> set[str]:
        # list existed storage credentials that is using service principal, capture the service principal's application_id
        application_ids = set()

        storage_credentials = self._ws.storage_credentials.list(max_results=0)

        if include_names:
            # we only check UC storage credentials listed in include_names
            for storage_credential in storage_credentials:
                if not storage_credential.azure_service_principal:
                    continue
                if storage_credential.name in include_names:
                    application_ids.add(storage_credential.azure_service_principal.application_id)
            logger.info(
                f"Found {len(application_ids)} distinct service principals already used in UC storage credentials listed in include_names"
            )
            return application_ids

        for storage_credential in storage_credentials:
            # only add service principal's application_id, ignore managed identity based storage_credential
            if storage_credential.azure_service_principal:
                application_ids.add(storage_credential.azure_service_principal.application_id)

        logger.info(f"Found {len(application_ids)} distinct service principals already used in UC storage credentials")
        return application_ids

    def create_with_client_secret(self, spn: ServicePrincipalMigrationInfo) -> StorageCredentialInfo:
        # this function should only be used to migrate service principal, fail the command here if
        # it's misused to create storage credential with managed identity
        assert spn.permission_mapping.directory_id is not None

        # prepare the storage credential properties
        name = spn.permission_mapping.principal
        service_principal = AzureServicePrincipal(
            spn.permission_mapping.directory_id,
            spn.permission_mapping.client_id,
            spn.client_secret,
        )
        comment = (
            f"Created by UCX during migration to UC using Azure Service Principal: {spn.permission_mapping.principal}"
        )

        # create the storage credential
        return self._ws.storage_credentials.create(
            name,
            azure_service_principal=service_principal,
            comment=comment,
            read_only=spn.permission_mapping.privilege == Privilege.READ_FILES.value,
        )

    def validate(self, storage_credential_info: StorageCredentialInfo, url: str) -> StorageCredentialValidationResult:
        try:
            validation = self._ws.storage_credentials.validate(
                storage_credential_name=storage_credential_info.name,
                url=url,
                read_only=storage_credential_info.read_only,
            )
        except InvalidParameterValue:
            logger.warning(
                "There is an existing external location overlaps with the prefix that is mapped to "
                "the service principal and used for validating the migrated storage credential. "
                "Skip the validation"
            )
            return StorageCredentialValidationResult.from_storage_credential_info(
                storage_credential_info,
                url,
                [
                    "The validation is skipped because an existing external location overlaps "
                    "with the location used for validation."
                ],
            )

        if not validation.results:
            return StorageCredentialValidationResult.from_storage_credential_info(
                storage_credential_info, url, ["Validation returned no results."]
            )

        failures = []
        for result in validation.results:
            if result.operation is None:
                continue
            if result.result == ValidationResultResult.FAIL:
                failures.append(f"{result.operation.value} validation failed with message: {result.message}")

        return StorageCredentialValidationResult.from_storage_credential_info(storage_credential_info, url, failures)


class ServicePrincipalMigration(SecretsMixin):

    def __init__(
        self,
        installation: Installation,
        ws: WorkspaceClient,
        resource_permissions: AzureResourcePermissions,
        service_principal_crawler: AzureServicePrincipalCrawler,
        storage_credential_manager: StorageCredentialManager,
    ):
        self._output_file = "azure_service_principal_migration_result.csv"
        self._installation = installation
        self._ws = ws
        self._resource_permissions = resource_permissions
        self._sp_crawler = service_principal_crawler
        self._storage_credential_manager = storage_credential_manager

    def _fetch_client_secret(self, sp_list: list[StoragePermissionMapping]) -> list[ServicePrincipalMigrationInfo]:
        # check AzureServicePrincipalInfo from AzureServicePrincipalCrawler, if AzureServicePrincipalInfo
        # has secret_scope and secret_key not empty, fetch the client_secret and put it to the client_secret field
        #
        # The input StoragePermissionMapping may have managed identity mixed in, we will ignore them for now, as
        # they won't have any client_secret, we will process managed identity in the future.

        # fetch client_secrets of crawled service principal, if any
        sp_info_with_client_secret: dict[str, str] = {}
        sp_infos = self._sp_crawler.snapshot()

        for sp_info in sp_infos:
            if not sp_info.secret_scope:
                continue
            if not sp_info.secret_key:
                continue

            secret_value = self._get_secret_if_exists(sp_info.secret_scope, sp_info.secret_key)

            if secret_value:
                sp_info_with_client_secret[sp_info.application_id] = secret_value
            else:
                logger.info(
                    f"Cannot fetch the service principal client_secret for {sp_info.application_id}. "
                    f"This service principal will be skipped for migration"
                )

        # update the list of ServicePrincipalMigrationInfo if client_secret is found
        sp_list_with_secret = []
        for spn in sp_list:
            if spn.client_id in sp_info_with_client_secret:
                sp_list_with_secret.append(
                    ServicePrincipalMigrationInfo(spn, sp_info_with_client_secret[spn.client_id])
                )
        return sp_list_with_secret

    def _print_action_plan(self, sp_list: list[StoragePermissionMapping]):
        # print action plan to console for customer to review.
        for spn in sp_list:
            logger.info(
                f"Service Principal name: {spn.principal}, "
                f"application_id: {spn.client_id}, "
                f"privilege {spn.privilege} "
                f"on location {spn.prefix}"
            )

    def _generate_migration_list(self, include_names: set[str] | None = None) -> list[ServicePrincipalMigrationInfo]:
        """
        Create the list of SP that need to be migrated, output an action plan as a csv file for users to confirm
        """
        # load sp list from azure_storage_account_info.csv
        permission_mappings = self._resource_permissions.load()
        # For now we only migrate Service Principal and ignore Managed Identity
        sp_list = [mapping for mapping in permission_mappings if mapping.type == "Application"]
        # list existed storage credentials
        sc_set = self._storage_credential_manager.list(include_names)
        # check if the sp is already used in UC storage credential
        filtered_sp_list = [sp for sp in sp_list if sp.client_id not in sc_set]
        # fetch sp client_secret if any
        sp_list_with_secret = self._fetch_client_secret(filtered_sp_list)

        # output the action plan for customer to confirm
        # but first make a copy of the list and strip out the client_secret
        sp_candidates = [sp.permission_mapping for sp in sp_list_with_secret]
        self._print_action_plan(sp_candidates)

        return sp_list_with_secret

    def save(self, migration_results: list[StorageCredentialValidationResult]) -> str:
        return self._installation.save(migration_results, filename=self._output_file)

    def _migrate_service_principals(
        self, sp_migration_infos: list[ServicePrincipalMigrationInfo]
    ) -> list[StorageCredentialValidationResult]:
        execution_result = []
        for spn in sp_migration_infos:
            if spn.permission_mapping.default_network_action != "Allow":
                logger.warning(
                    f"Service principal '{spn.permission_mapping.principal}' accesses storage account "
                    f"'{spn.permission_mapping.prefix}' with non-Allow network configuration"
                )
            try:

                storage_credential_info = self._storage_credential_manager.create_with_client_secret(spn)
                validation_results = self._storage_credential_manager.validate(
                    storage_credential_info,
                    spn.permission_mapping.prefix,
                )
                execution_result.append(validation_results)
            except (PermissionDenied, NotFound, BadRequest) as e:
                logger.error(
                    "Error creating storage credentials. Please check error message and try again."
                    "Rolling back storage credential that are created."
                )
                self._resource_permissions.delete_storage_credential(execution_result)
                raise e
        return execution_result

    def _create_storage_credentials_for_storage_accounts(
        self,
    ) -> list[StorageCredentialValidationResult]:
        access_connectors = self._resource_permissions.create_access_connectors_for_storage_accounts()

        execution_results: list[StorageCredentialValidationResult] = []
        for access_connector, url in access_connectors:
            storage_credential_info = self._ws.storage_credentials.create(
                access_connector.name,
                azure_managed_identity=AzureManagedIdentityRequest(str(access_connector.id)),
                comment="Created by UCX",
                read_only=False,  # Access connectors get "STORAGE_BLOB_DATA_CONTRIBUTOR" permissions
            )
            try:
                validation_results = self._storage_credential_manager.validate(storage_credential_info, url)
            except BadRequest:
                logger.warning(f"Could not validate storage credential {storage_credential_info.name} for url {url}")
            except (PermissionDenied, NotFound):
                logger.error(
                    "Error creating access connectors. Please review the error message and resolve the issue before "
                    "trying again. "
                    "Removing successfully created storage credentials."
                )
                self._resource_permissions.delete_storage_credential(*execution_results)
                delete_access_connectors = [access_connector for access_connector, _ in access_connectors]
                self._resource_permissions.delete_access_connectors(*delete_access_connectors)
            else:
                execution_results.append(validation_results)

        return execution_results

    def run(self, prompts: Prompts, include_names: set[str] | None = None) -> list[StorageCredentialValidationResult]:
        plan_confirmed = prompts.confirm(
            "[RECOMMENDED] Please confirm to create an access connector with a managed identity for each storage "
            "account."
        )
        ac_results: list[StorageCredentialValidationResult] = []
        if plan_confirmed:
            ac_results = self._create_storage_credentials_for_storage_accounts()

        sp_migration_infos = self._generate_migration_list(include_names)
        if any(spn.permission_mapping.default_network_action != "Allow" for spn in sp_migration_infos):
            logger.warning(
                "At least one Azure Service Principal accesses a storage account with non-Allow default network "
                "configuration, which might cause connectivity issues. We recommend using Databricks Access "
                "Connectors instead"
            )
        plan_confirmed = prompts.confirm(
            "Above Azure Service Principals will be migrated to UC storage credentials, please review and confirm."
        )
        sp_results = []
        if plan_confirmed:
            sp_results = self._migrate_service_principals(sp_migration_infos)

        execution_results = ac_results + sp_results
        if execution_results:
            results_file = self.save(execution_results)
            logger.info(
                "Completed migration from Azure Service Principal to UC Storage credentials "
                "and creation of UC Storage credentials for storage access with Databricks Access Connectors. "
                f"Please check {results_file} for validation results"
            )
        else:
            logger.info(
                "No UC Storage credentials created during Azure Service Principal migration "
                "nor for storage access with Databricks Access Connectors."
            )

        return execution_results



import logging

from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import AccountClient
from databricks.sdk.errors import NotFound
from databricks.sdk.service.settings import DefaultNamespaceSetting, StringMessage

logger = logging.getLogger(__name__)


class AccountMetastores:
    def __init__(
        self,
        account_client: AccountClient,
    ):
        self._ac = account_client

    def show_all_metastores(self, workspace_id: str | None = None):
        location = None
        if workspace_id:
            logger.info(f"Workspace ID: {workspace_id}")
            location = self._get_region(int(workspace_id))
        logger.info("Matching metastores are:")
        for metastore in self._get_all_metastores(location).keys():
            logger.info(metastore)

    def assign_metastore(
        self,
        prompts: Prompts,
        workspace_id: int,
        *,
        metastore_id: str | None = None,
        default_catalog: str | None = None,
    ):
        if metastore_id is None:
            # search for all matching metastores
            metastore_choices = self._get_all_metastores(self._get_region(workspace_id))
            if len(metastore_choices) == 0:
                raise ValueError(f"No matching metastore found for workspace: {workspace_id}")
            # if there are multiple matches, prompt users to select one
            if len(metastore_choices) > 1:
                metastore_id = prompts.choice_from_dict(
                    "Multiple metastores found, please select one:", metastore_choices
                )
            else:
                metastore_id = list(metastore_choices.values())[0]
        self._ac.metastore_assignments.create(workspace_id, metastore_id)
        # set the default catalog using the default_namespace setting API
        if default_catalog is not None:
            self._set_default_catalog(workspace_id, default_catalog)

    def _get_region(self, workspace_id: int) -> str | None:
        workspace = self._ac.workspaces.get(workspace_id)
        if self._ac.config.is_aws:
            return workspace.aws_region
        return workspace.location

    def _get_all_workspaces(self) -> dict[str, int]:
        output = dict[str, int]()
        for workspace in self._ac.workspaces.list():
            if workspace.workspace_id:
                output[f"{workspace.workspace_name} - {workspace.workspace_id}"] = workspace.workspace_id
        return dict(sorted(output.items()))

    def _get_all_metastores(self, location: str | None = None) -> dict[str, str]:
        output = dict[str, str]()
        for metastore in self._ac.metastores.list():
            if location is None or metastore.region == location:
                output[f"{metastore.name} - {metastore.metastore_id}"] = str(metastore.metastore_id)
        return dict(sorted(output.items()))

    def _set_default_catalog(self, workspace_id: int, default_catalog: str):
        if default_catalog == "":
            return
        workspace = self._ac.workspaces.get(int(workspace_id))
        default_namespace = self._ac.get_workspace_client(workspace).settings.default_namespace
        # needs to get the etag first, before patching the setting
        try:
            etag = default_namespace.get().etag
        except NotFound as err:
            # if not found, the etag is returned in the header
            etag = err.details[0].metadata.get("etag")
        default_namespace.update(
            allow_missing=True,
            field_mask="namespace.value",
            setting=DefaultNamespaceSetting(etag=etag, namespace=StringMessage(default_catalog)),
        )



import logging
from collections import defaultdict
from collections.abc import Callable, Iterable
from dataclasses import dataclass, replace
from functools import partial, cached_property

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.parallel import ManyError, Threads
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import ResourceDoesNotExist, NotFound, DatabricksError
from databricks.sdk.service.catalog import (
    ExternalLocationInfo,
    SchemaInfo,
    TableInfo,
    Privilege,
    PermissionsChange,
    SecurableType,
)
from databricks.sdk.service.compute import ClusterSource, DataSecurityMode

from databricks.labs.ucx.account.workspaces import WorkspaceInfo
from databricks.labs.ucx.assessment.aws import AWSRoleAction
from databricks.labs.ucx.assessment.azure import (
    AzureServicePrincipalCrawler,
    AzureServicePrincipalInfo,
)
from databricks.labs.ucx.aws.access import AWSResourcePermissions
from databricks.labs.ucx.azure.access import (
    AzureResourcePermissions,
    StoragePermissionMapping,
)
from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.locations import (
    ExternalLocations,
    Mount,
    Mounts,
)
from databricks.labs.ucx.hive_metastore.mapping import TableToMigrate, Rule
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationStatusRefresher
from databricks.labs.ucx.hive_metastore.tables import Table, TablesCrawler
from databricks.labs.ucx.hive_metastore.udfs import UdfsCrawler
from databricks.labs.ucx.workspace_access.groups import GroupManager

logger = logging.getLogger(__name__)


@dataclass
class ComputeLocations:
    compute_id: str
    locations: dict
    compute_type: str


@dataclass(frozen=True)
class Grant:
    principal: str
    action_type: str
    catalog: str | None = None
    database: str | None = None
    table: str | None = None
    view: str | None = None
    udf: str | None = None
    any_file: bool = False
    anonymous_function: bool = False

    @staticmethod
    def type_and_key(
        *,
        catalog: str | None = None,
        database: str | None = None,
        table: str | None = None,
        view: str | None = None,
        udf: str | None = None,
        any_file: bool = False,
        anonymous_function: bool = False,
    ) -> tuple[str, str]:
        if table is not None:
            catalog = "hive_metastore" if catalog is None else catalog
            database = "default" if database is None else database
            return "TABLE", f"{catalog}.{database}.{table}"
        if view is not None:
            catalog = "hive_metastore" if catalog is None else catalog
            database = "default" if database is None else database
            return "VIEW", f"{catalog}.{database}.{view}"
        if udf is not None:
            catalog = "hive_metastore" if catalog is None else catalog
            database = "default" if database is None else database
            return "FUNCTION", f"{catalog}.{database}.{udf}"
        if database is not None:
            catalog = "hive_metastore" if catalog is None else catalog
            return "DATABASE", f"{catalog}.{database}"
        if any_file:
            return "ANY FILE", ""
        if anonymous_function:
            return "ANONYMOUS FUNCTION", ""
        # Must come last, as it has the lowest priority here but is a required parameter
        if catalog is not None:
            return "CATALOG", catalog
        msg = (
            f"invalid grant keys: catalog={catalog}, database={database}, view={view}, udf={udf}"
            f"any_file={any_file}, anonymous_function={anonymous_function}"
        )
        raise ValueError(msg)

    @property
    def object_key(self) -> str:
        _, key = self.this_type_and_key()
        return key.lower()

    def this_type_and_key(self):
        return self.type_and_key(
            catalog=self.catalog,
            database=self.database,
            table=self.table,
            view=self.view,
            udf=self.udf,
            any_file=self.any_file,
            anonymous_function=self.anonymous_function,
        )

    def hive_grant_sql(self) -> list[str]:
        object_type, object_key = self.this_type_and_key()
        # See https://docs.databricks.com/en/sql/language-manual/security-grant.html
        statements = []
        actions = self.action_type.split(", ")
        deny_actions = [action for action in actions if "DENIED" in action]
        grant_actions = [action for action in actions if "DENIED" not in action]
        if "OWN" in grant_actions:
            grant_actions.remove("OWN")
            statements.append(self._set_owner_sql(object_type, object_key))
        if grant_actions:
            statements.append(self._apply_grant_sql(", ".join(grant_actions), object_type, object_key))
        if deny_actions:
            statements.append(self._apply_grant_sql(", ".join(deny_actions), object_type, object_key))
        return statements

    def hive_revoke_sql(self) -> str:
        object_type, object_key = self.this_type_and_key()
        return f"REVOKE {self.action_type} ON {object_type} {escape_sql_identifier(object_key)} FROM `{self.principal}`"

    def _set_owner_sql(self, object_type, object_key):
        return f"ALTER {object_type} {escape_sql_identifier(object_key)} OWNER TO `{self.principal}`"

    def _apply_grant_sql(self, action_type, object_type, object_key):
        if "DENIED" in action_type:
            action_type = action_type.replace("DENIED_", "")
            # need to wrap the action type in backticks to avoid syntax errors with DENY SELECT
            return f"DENY `{action_type}` ON {object_type} {escape_sql_identifier(object_key)} TO `{self.principal}`"
        return f"GRANT {action_type} ON {object_type} {escape_sql_identifier(object_key)} TO `{self.principal}`"

    def _uc_action(self, action_type):
        def inner(object_type, object_key):
            return self._apply_grant_sql(action_type, object_type, object_key)

        return inner

    def uc_grant_sql(self, object_type: str | None = None, object_key: str | None = None) -> str | None:
        """Get SQL translated SQL statement for granting similar permissions in UC.

        If there's no UC equivalent, returns None. This can also be the case for missing mapping.
        """

        # TODO: verify and complete the mapping
        # See: https://docs.databricks.com/sql/language-manual/sql-ref-privileges-hms.html
        # See: https://docs.databricks.com/data-governance/unity-catalog/manage-privileges/ownership.html
        # See: https://docs.databricks.com/data-governance/unity-catalog/manage-privileges/privileges.html
        if object_type is None:
            object_type, object_key = self.this_type_and_key()
        hive_to_uc = {
            ("FUNCTION", "SELECT"): self._uc_action("EXECUTE"),
            ("TABLE", "SELECT"): self._uc_action("SELECT"),
            ("TABLE", "MODIFY"): self._uc_action("MODIFY"),
            ("TABLE", "ALL PRIVILEGES"): self._uc_action("ALL PRIVILEGES"),
            ("TABLE", "OWN"): self._set_owner_sql,
            ("VIEW", "SELECT"): self._uc_action("SELECT"),
            ("VIEW", "OWN"): self._set_owner_sql,
            ("DATABASE", "USAGE"): self._uc_action("USE SCHEMA"),
            ("DATABASE", "CREATE"): self._uc_action("CREATE TABLE"),
            ("DATABASE", "CREATE_NAMED_FUNCTION"): self._uc_action("CREATE FUNCTION"),
            ("DATABASE", "SELECT"): self._uc_action("SELECT"),
            ("DATABASE", "MODIFY"): self._uc_action("MODIFY"),
            ("DATABASE", "OWN"): self._set_owner_sql,
            ("CATALOG", "OWN"): self._set_owner_sql,
            ("CATALOG", "USAGE"): self._uc_action("USE CATALOG"),
        }
        make_query = hive_to_uc.get((object_type, self.action_type), None)
        if make_query is None:
            # unknown mapping or ignore
            return None
        return make_query(object_type, object_key)


CLUSTER_WITHOUT_ACL_FRAGMENT = "Table Access Control is not enabled on this cluster"


class GrantsCrawler(CrawlerBase[Grant]):
    """Crawler that captures access controls that relate to data and other securable objects."""

    def __init__(self, tc: TablesCrawler, udf: UdfsCrawler, include_databases: list[str] | None = None):
        assert tc._backend == udf._backend
        assert tc._catalog == udf._catalog
        assert tc._schema == udf._schema
        super().__init__(tc._backend, tc._catalog, tc._schema, "grants", Grant)
        self._tc = tc
        self._udf = udf
        self._include_databases = include_databases

    def snapshot(self, *, force_refresh: bool = False) -> Iterable[Grant]:
        try:
            return super().snapshot(force_refresh=force_refresh)
        except Exception as e:  # pylint: disable=broad-exception-caught
            log_fn = logger.warning if CLUSTER_WITHOUT_ACL_FRAGMENT in repr(e) else logger.error
            log_fn(f"Couldn't fetch grants snapshot: {e}")
            return []

    def _try_fetch(self):
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield Grant(*row)

    def _crawl(self) -> Iterable[Grant]:
        """
        Crawls and lists grants for all databases, tables,  views, udfs, any file
        and anonymous function within hive_metastore.

        Returns:
            list[Grant]: A list of Grant objects representing the listed grants.

        Behavior:
        - Constructs a list of tasks to fetch grants using the `_grants` method, including both database-wide and
          table/view-specific grants.
        - Iterates through tables in the specified database using the `_tc.snapshot` method.
        - For each table, adds tasks to fetch grants for the table or its view, depending on the kind of the table.
        - Iterates through udfs in the specified database using the `_udf.snapshot` method.
        - For each udf, adds tasks to fetch grants for the udf.
        - Executes the tasks concurrently using Threads.gather.
        - Flattens the list of retrieved grant lists into a single list of Grant objects.

        Note:
        - The method assumes that the `_grants` method fetches grants based on the provided parameters (catalog,
          database, table, view, udfs, any file, anonymous function).

        Returns:
        list[Grant]: A list of Grant objects representing the grants found in hive_metastore.
        """
        catalog = "hive_metastore"
        tasks = [
            partial(self.grants, catalog=catalog),
            # Scanning ANY FILE and ANONYMOUS FUNCTION grants
            partial(self.grants, catalog=catalog, any_file=True),
            partial(self.grants, catalog=catalog, anonymous_function=True),
        ]
        if not self._include_databases:
            # scan all databases, even empty ones
            for row in self._fetch(f"SHOW DATABASES FROM {escape_sql_identifier(catalog)}"):
                tasks.append(partial(self.grants, catalog=catalog, database=row.databaseName))
        else:
            for database in self._include_databases:
                tasks.append(partial(self.grants, catalog=catalog, database=database))
        for table in self._tc.snapshot():
            fn = partial(self.grants, catalog=catalog, database=table.database)
            # Views are treated as a type of table and enumerated by the table crawler.
            if table.view_text is None:
                task = partial(fn, table=table.name)
            else:
                task = partial(fn, view=table.name)
            tasks.append(task)
        for udf in self._udf.snapshot():
            fn = partial(self.grants, catalog=catalog, database=udf.database)
            tasks.append(partial(fn, udf=udf.name))
        catalog_grants, errors = Threads.gather(f"listing grants for {catalog}", tasks)
        if len(errors) > 0:
            raise ManyError(errors)
        return [grant for grants in catalog_grants for grant in grants]

    def for_table_info(self, table: TableInfo):
        # TODO: it does not work yet for views
        principal_permissions = defaultdict(set)
        for grant in self.grants(catalog=table.catalog_name, database=table.schema_name, table=table.name):
            principal_permissions[grant.principal].add(grant.action_type)
        return principal_permissions

    def for_schema_info(self, schema: SchemaInfo):
        principal_permissions = defaultdict(set)
        for grant in self.grants(catalog=schema.catalog_name, database=schema.name):
            principal_permissions[grant.principal].add(grant.action_type)
        return principal_permissions

    # The reported ObjectType for grants doesn't necessary match the type of grants we asked for. This maps
    # those that aren't themselves.
    _grants_reported_as = {
        # SHOW type: RESULT type
        "SCHEMA": "DATABASE",
        "CATALOG": "CATALOG$",
        "VIEW": "TABLE",
        "ANY FILE": "ANY_FILE",
        "ANONYMOUS FUNCTION": "ANONYMOUS_FUNCTION",
    }

    def grants(
        self,
        *,
        catalog: str | None = None,
        database: str | None = None,
        table: str | None = None,
        view: str | None = None,
        udf: str | None = None,
        any_file: bool = False,
        anonymous_function: bool = False,
    ) -> list[Grant]:
        """
        Fetches and yields grant information for the specified database objects.

        Keyword Args:
            catalog (str): The catalog name (optional).
            database (str | None): The database name (optional).
            table (str | None): The table name (optional).
            view (str | None): The view name (optional).
            udf (str | None): The udf name (optional).
            any_file (bool): Whether to include any file grants (optional).
            anonymous_function (bool): Whether to include anonymous function grants (optional).

        Behavior:
        - Normalizes the provided parameters and constructs an object type and key using
          the `Grant.type_and_key` method.
        - Iterates through rows fetched using the `_fetch` method by executing a SQL query
          to retrieve grant information.
        - For each fetched row, extracts the principal, action type, and object type.
        - Normalizes the object type and filters grants based on the provided object type
          and the fetched object type.
        - Yields a Grant object representing the fetched grant information.

        Note:
        - The method fetches and yields grants based on the provided parameters and
          the available grant information in the database.

        Returns:
        Iterator[Grant]: An iterator of Grant objects representing the fetched grant information.
        """
        on_type, key = Grant.type_and_key(
            catalog=self._try_valid(catalog),
            database=self._try_valid(database),
            table=self._try_valid(table),
            view=self._try_valid(view),
            udf=self._try_valid(udf),
            any_file=any_file,
            anonymous_function=anonymous_function,
        )
        expected_grant_object_type = self._grants_reported_as.get(on_type, on_type)
        grants = []
        try:
            for row in self._fetch(f"SHOW GRANTS ON {on_type} {escape_sql_identifier(key)}"):
                (principal, action_type, object_type, _) = row
                # Sometimes we get grants for other objects we didn't ask for. For example, listing DATABASE grants
                # may also enumerate grants on the associated CATALOG.
                if object_type != expected_grant_object_type:
                    continue
                # we have to return concrete list, as with yield we're executing
                # everything on the main thread.
                grant = Grant(
                    principal=principal,
                    action_type=action_type,
                    table=table,
                    view=view,
                    udf=udf,
                    database=database,
                    catalog=catalog,
                    any_file=any_file,
                    anonymous_function=anonymous_function,
                )
                grants.append(grant)
            return grants
        except NotFound:
            # This makes the integration test more robust as many test schemas are being created and deleted quickly.
            logger.warning(f"Schema {catalog}.{database} no longer existed")
            return []
        except Exception as e:  # pylint: disable=broad-exception-caught
            log_fn = logger.warning if CLUSTER_WITHOUT_ACL_FRAGMENT in repr(e) else logger.error
            log_fn(f"Couldn't fetch grants for object {on_type} {key}: {e}")
            return []


class AwsACL:
    def __init__(
        self,
        ws: WorkspaceClient,
        backend: SqlBackend,
        installation: Installation,
    ):
        self._backend = backend
        self._ws = ws
        self._installation = installation

    def _get_cluster_to_instance_profile_mapping(self) -> dict[str, str]:
        # this function gives a mapping between an interactive cluster and the instance profile used by it
        # either directly or through a cluster policy.
        cluster_instance_profiles = {}

        for cluster in self._ws.clusters.list():
            if (
                cluster.cluster_id is None
                or cluster.cluster_source == ClusterSource.JOB
                or (cluster.data_security_mode not in [DataSecurityMode.LEGACY_SINGLE_USER, DataSecurityMode.NONE])
            ):
                continue
            if cluster.aws_attributes is None:
                continue
            if cluster.aws_attributes.instance_profile_arn is not None:
                role_name = cluster.aws_attributes.instance_profile_arn
                assert role_name is not None
                cluster_instance_profiles[cluster.cluster_id] = role_name
                continue

        return cluster_instance_profiles

    def _update_warehouse_to_instance_profile_mapping(
        self,
    ) -> dict[str, str]:
        warehouse_instance_profiles = {}
        sql_config = self._ws.warehouses.get_workspace_warehouse_config()
        if sql_config.instance_profile_arn is not None:
            role_name = sql_config.instance_profile_arn
            for warehouse in self._ws.warehouses.list():
                if warehouse.id is not None:
                    warehouse_instance_profiles[warehouse.id] = role_name
        return warehouse_instance_profiles

    def get_eligible_locations_principals(self) -> list[ComputeLocations]:
        cluster_instance_profiles = self._get_cluster_to_instance_profile_mapping()
        warehouse_instance_profiles = self._update_warehouse_to_instance_profile_mapping()
        compute_locations = []
        if len(cluster_instance_profiles) == 0 and len(warehouse_instance_profiles) == 0:
            # if there are no interactive clusters or warehouse with instance profile , then return empty grants
            logger.info("No interactive cluster or sql warehouse found with instance profiles configured")
            return []
        external_locations = list(self._ws.external_locations.list())
        if len(external_locations) == 0:
            # if there are no external locations, then throw an error to run migrate_locations cli command
            msg = (
                "No external location found, If hive metastore tables are created in external storage, "
                "ensure migrate-locations cli cmd is run to create the required locations."
            )
            logger.error(msg)
            raise ResourceDoesNotExist(msg) from None

        permission_mappings = self._installation.load(
            list[AWSRoleAction],
            filename=AWSResourcePermissions.INSTANCE_PROFILES_FILE_NAME,
        )
        if len(permission_mappings) == 0:
            # if permission mapping is empty, raise an error to run principal_prefix cmd
            msg = (
                "No instance profile permission file found. Please ensure principal-prefix-access cli "
                "cmd is run to create the instance profile permission file."
            )
            logger.error(msg)
            raise ResourceDoesNotExist(msg) from None

        for cluster_id, role_compute in cluster_instance_profiles.items():
            eligible_locations = self._get_external_locations(role_compute, external_locations, permission_mappings)
            if len(eligible_locations) == 0:
                continue
            compute_locations.append(ComputeLocations(cluster_id, eligible_locations, "clusters"))
        for warehouse_id, role_compute in warehouse_instance_profiles.items():
            eligible_locations = self._get_external_locations(role_compute, external_locations, permission_mappings)
            if len(eligible_locations) == 0:
                continue
            compute_locations.append(ComputeLocations(warehouse_id, eligible_locations, "warehouses"))
        return compute_locations

    @staticmethod
    def _get_external_locations(
        role_name: str,
        external_locations: list[ExternalLocationInfo],
        permission_mappings: list[AWSRoleAction],
    ) -> dict[str, str]:
        matching_location = {}
        for location in external_locations:
            if location.url is None:
                continue
            for permission_mapping in permission_mappings:
                # check if resource_path contains "*"
                if permission_mapping.resource_path.find('*') > 1:
                    resource_url = permission_mapping.resource_path[: permission_mapping.resource_path.index('*') - 1]
                else:
                    resource_url = permission_mapping.resource_path
                if location.url.startswith(resource_url) and permission_mapping.role_arn == role_name:
                    matching_location[location.url] = permission_mapping.privilege
        return matching_location


class AzureACL:
    def __init__(
        self,
        ws: WorkspaceClient,
        backend: SqlBackend,
        spn_crawler: AzureServicePrincipalCrawler,
        installation: Installation,
    ):
        self._backend = backend
        self._ws = ws
        self._spn_crawler = spn_crawler
        self._installation = installation

    def get_eligible_locations_principals(self) -> list[ComputeLocations]:
        compute_locations = []
        spn_cluster_mapping = self._spn_crawler.get_cluster_to_storage_mapping()
        spn_warehouse_mapping = self._spn_crawler.get_warehouse_to_storage_mapping()
        if len(spn_cluster_mapping) == 0 and len(spn_warehouse_mapping) == 0:
            # if there are no interactive clusters , then return empty grants
            logger.info("No interactive cluster found with spn configured")
            return []
        external_locations = list(self._ws.external_locations.list())
        if len(external_locations) == 0:
            # if there are no external locations, then throw an error to run migrate_locations cli command
            msg = (
                "No external location found, If hive metastore tables are created in external storage, "
                "ensure migrate-locations cli cmd is run to create the required locations."
            )
            logger.error(msg)
            raise ResourceDoesNotExist(msg) from None

        permission_mappings = self._installation.load(
            list[StoragePermissionMapping],
            filename=AzureResourcePermissions.FILENAME,
        )
        if len(permission_mappings) == 0:
            # if permission mapping is empty, raise an error to run principal_prefix cmd
            msg = (
                "No storage permission file found. Please ensure principal-prefix-access cli "
                "cmd is run to create the access permission file."
            )
            logger.error(msg)
            raise ResourceDoesNotExist(msg) from None

        for cluster_spn in spn_cluster_mapping:
            eligible_locations = {}
            for spn in cluster_spn.spn_info:
                eligible_locations.update(self._get_external_locations(spn, external_locations, permission_mappings))
            compute_locations.append(ComputeLocations(cluster_spn.cluster_id, eligible_locations, "clusters"))

        for warehouse_spn in spn_warehouse_mapping:
            eligible_locations = {}
            for spn in warehouse_spn.spn_info:
                eligible_locations.update(self._get_external_locations(spn, external_locations, permission_mappings))
            compute_locations.append(ComputeLocations(warehouse_spn.cluster_id, eligible_locations, "warehouses"))

        return compute_locations

    def _get_external_locations(
        self,
        spn: AzureServicePrincipalInfo,
        external_locations: list[ExternalLocationInfo],
        permission_mappings: list[StoragePermissionMapping],
    ) -> dict[str, str]:
        matching_location = {}
        for location in external_locations:
            if location.url is None:
                continue
            for permission_mapping in permission_mappings:
                prefix = permission_mapping.prefix
                if (
                    location.url.startswith(permission_mapping.prefix)
                    and permission_mapping.client_id == spn.application_id
                    and spn.storage_account is not None
                    # check for storage account name starting after @ in the prefix url
                    and prefix[prefix.index('@') + 1 :].startswith(spn.storage_account)
                ):
                    matching_location[location.url] = permission_mapping.privilege
        return matching_location


class PrincipalACL:
    def __init__(
        self,
        ws: WorkspaceClient,
        backend: SqlBackend,
        installation: Installation,
        tables_crawler: TablesCrawler,
        mounts_crawler: Mounts,
        cluster_locations: Callable[[], list[ComputeLocations]],
    ):
        self._backend = backend
        self._ws = ws
        self._installation = installation
        self._tables_crawler = tables_crawler
        self._mounts_crawler = mounts_crawler
        self._compute_locations = cluster_locations

    def get_interactive_cluster_grants(self) -> list[Grant]:
        tables = list(self._tables_crawler.snapshot())
        mounts = list(self._mounts_crawler.snapshot())
        grants: set[Grant] = set()

        for compute_location in self._compute_locations():
            principals = self._get_cluster_principal_mapping(compute_location.compute_id, compute_location.compute_type)
            if len(principals) == 0:
                continue
            cluster_usage = self._get_grants(compute_location.locations, principals, tables, mounts)
            grants.update(cluster_usage)
        return list(grants)

    def _get_privilege(self, table: Table, locations: dict[str, str], mounts: list[Mount]) -> str | None:
        if table.view_text is not None:
            # return nothing for view so that it goes to the separate view logic
            return None
        if table.location is None:
            return None
        if table.location.startswith('dbfs:/mnt') or table.location.startswith('/dbfs/mnt'):
            mount_location = ExternalLocations.resolve_mount(table.location, mounts)
            for loc, privilege in locations.items():
                if loc is not None and mount_location.startswith(loc):
                    return privilege
            return None
        if table.location.startswith('dbfs:/') or table.location.startswith('/dbfs/'):
            return "WRITE_FILES"

        for loc, privilege in locations.items():
            if loc is not None and table.location.startswith(loc):
                return privilege
        return None

    def _get_database_grants(self, tables: list[Table], principals: list[str]) -> list[Grant]:
        databases = {table.database for table in tables}
        grants = []
        for database in databases:
            for principal in principals:
                grant = Grant(principal, "USAGE", "hive_metastore", database)
                grants.append(grant)
        return grants

    def _get_grants(
        self, locations: dict[str, str], principals: list[str], tables: list[Table], mounts: list[Mount]
    ) -> list[Grant]:
        grants = []
        filtered_tables = []
        for table in tables:
            privilege = self._get_privilege(table, locations, mounts)
            if privilege == "READ_FILES":
                for principal in principals:
                    grants.append(Grant(principal, "SELECT", table.catalog, table.database, table.name))
                filtered_tables.append(table)
                continue
            if privilege == "WRITE_FILES":
                for principal in principals:
                    grants.append(Grant(principal, "ALL PRIVILEGES", table.catalog, table.database, table.name))
                filtered_tables.append(table)
                continue
            if table.view_text is not None:
                for principal in principals:
                    grants.append(Grant(principal, "ALL PRIVILEGES", table.catalog, table.database, view=table.name))
                filtered_tables.append(table)
        database_grants = self._get_database_grants(filtered_tables, principals)
        grants.extend(database_grants)
        return grants

    def _get_cluster_principal_mapping(self, cluster_id: str, object_type: str) -> list[str]:
        # gets all the users,groups,spn which have access to the clusters and returns a dataclass of that mapping
        principal_list = []
        try:
            cluster_permission = self._ws.permissions.get(object_type, cluster_id)
        except ResourceDoesNotExist:
            return []
        if cluster_permission.access_control_list is None:
            return []
        for acl in cluster_permission.access_control_list:
            if acl.user_name is not None:
                principal_list.append(acl.user_name)
            if acl.group_name is not None:
                if acl.group_name == "admins":
                    continue
                principal_list.append(acl.group_name)
            if acl.service_principal_name is not None:
                principal_list.append(acl.service_principal_name)
        return principal_list

    def apply_location_acl(self):
        """
        Check the interactive cluster and the principals mapped to it
        identifies the spn or instance profile configured for the interactive cluster
        identifies any location the spn/instance profile have access to (read or write)
        applies create_external_table, create_external_volume and read_files permission for all location
        to the principal
        """
        logger.info(
            "Applying permission for external location (CREATE EXTERNAL TABLE, "
            "CREATE EXTERNAL VOLUME and READ_FILES for existing eligible interactive cluster users"
        )
        # get the eligible location mapped for each interactive cluster
        for compute_location in self._compute_locations():
            # get interactive cluster users
            principals = self._get_cluster_principal_mapping(compute_location.compute_id, compute_location.compute_type)
            if len(principals) == 0:
                continue
            for location_url in compute_location.locations.keys():
                # get the location name for the given url
                location_name = self._get_location_name(location_url)
                if location_name is None:
                    continue
                self._update_location_permissions(location_name, principals)
        logger.info("Applied all the permission on external location")

    def _update_location_permissions(self, location_name: str, principals: list[str]):
        permissions = [Privilege.CREATE_EXTERNAL_TABLE, Privilege.CREATE_EXTERNAL_VOLUME, Privilege.READ_FILES]
        changes = [PermissionsChange(add=permissions, principal=principal) for principal in principals]
        self._ws.grants.update(
            SecurableType.EXTERNAL_LOCATION,
            location_name,
            changes=changes,
        )

    def _get_location_name(self, location_url: str):
        for location in self._ws.external_locations.list():
            if location.url == location_url:
                return location.name
        return None


class MigrateGrants:
    def __init__(
        self,
        sql_backend: SqlBackend,
        group_manager: GroupManager,
        grant_loaders: list[Callable[[], Iterable[Grant]]],
    ):
        self._sql_backend = sql_backend
        self._group_manager = group_manager
        self._grant_loaders = grant_loaders

    def apply(self, src: Table, uc_table_key: str) -> bool:
        for grant in self._match_grants(src):
            acl_migrate_sql = grant.uc_grant_sql(src.kind, uc_table_key)
            if acl_migrate_sql is None:
                logger.warning(
                    f"failed-to-migrate: Hive metastore grant '{grant.action_type}' cannot be mapped to UC grant for "
                    f"{src.kind} '{uc_table_key}'. Skipping."
                )
                continue
            logger.debug(f"Migrating acls on {uc_table_key} using SQL query: {acl_migrate_sql}")
            try:
                self._sql_backend.execute(acl_migrate_sql)
            except DatabricksError as e:
                logger.warning(f"failed-to-migrate: Failed to migrate ACL for {src.key} to {uc_table_key}: {e}")
        return True

    @cached_property
    def _workspace_to_account_group_names(self) -> dict[str, str]:
        return {g.name_in_workspace: g.name_in_account for g in self._group_manager.snapshot()}

    @cached_property
    def _grants(self) -> list[Grant]:
        grants = []
        for loader in self._grant_loaders:
            for grant in loader():
                grants.append(grant)
        return grants

    def _match_grants(self, table: Table) -> list[Grant]:
        matched_grants = []
        for grant in self._grants:
            if grant.database != table.database:
                continue
            if table.name not in (grant.table, grant.view):
                continue
            grant = self._replace_account_group(grant)
            matched_grants.append(grant)
        return matched_grants

    def _replace_account_group(self, grant: Grant) -> Grant:
        target_principal = self._workspace_to_account_group_names.get(grant.principal)
        if not target_principal:
            return grant
        return replace(grant, principal=target_principal)


class ACLMigrator:
    def __init__(
        self,
        tables_crawler: TablesCrawler,
        workspace_info: WorkspaceInfo,
        migration_status_refresher: TableMigrationStatusRefresher,
        migrate_grants: MigrateGrants,
    ):
        self._table_crawler = tables_crawler
        self._workspace_info = workspace_info
        self._migration_status_refresher = migration_status_refresher
        self._migrate_grants = migrate_grants

    def migrate_acls(self, *, target_catalog: str | None = None, hms_fed: bool = False) -> None:
        workspace_name = self._workspace_info.current()
        tables = list(self._table_crawler.snapshot())
        if not tables:
            logger.info("No tables found to acl")
            return
        if hms_fed:
            tables_to_migrate = self._get_hms_fed_tables(tables, target_catalog if target_catalog else workspace_name)
        else:
            tables_to_migrate = self._get_migrated_tables(tables)
        self._migrate_acls(tables_to_migrate)

    def _get_migrated_tables(self, tables: list[Table]) -> list[TableToMigrate]:
        # gets all the migrated table to apply ACLs to
        tables_to_migrate = []
        seen_tables = self._migration_status_refresher.get_seen_tables()
        reverse_seen_tables = {v: k for k, v in seen_tables.items()}
        for table in tables:
            if table.key not in reverse_seen_tables:
                logger.warning(f"Table {table.key} not found in migration status. Skipping.")
                continue
            dst_table_parts = reverse_seen_tables[table.key].split(".")
            if len(dst_table_parts) != 3:
                logger.warning(
                    f"Invalid table name {reverse_seen_tables[table.key]} found in migration status. Skipping."
                )
                continue
            rule = Rule(
                self._workspace_info.current(),
                dst_table_parts[0],
                table.database,
                dst_table_parts[1],
                table.name,
                dst_table_parts[2],
            )
            table_to_migrate = TableToMigrate(table, rule)
            tables_to_migrate.append(table_to_migrate)
        return tables_to_migrate

    def _get_hms_fed_tables(self, tables: list[Table], target_catalog) -> list[TableToMigrate]:
        # if it is hms_fed acl migration, migrate all the acls for tables in the provided catalog
        tables_to_migrate = []
        for table in tables:
            rule = Rule(
                self._workspace_info.current(),
                target_catalog,
                table.database,
                table.database,
                table.name,
                table.name,
            )
            table_to_migrate = TableToMigrate(table, rule)
            tables_to_migrate.append(table_to_migrate)
        return tables_to_migrate

    def _migrate_acls(self, tables_in_scope: list[TableToMigrate]) -> None:
        tasks = []
        for table in tables_in_scope:
            tasks.append(partial(self._migrate_grants.apply, table.src, table.rule.as_uc_table_key))
        Threads.strict("migrate grants", tasks)

    def _is_migrated(self, schema: str, table: str) -> bool:
        index = self._migration_status_refresher.index()
        return index.is_migrated(schema, table)



from __future__ import annotations

import logging
from collections.abc import Sequence, Iterable

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk.errors import DatabricksError

from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.source_code.base import UsedTable

logger = logging.getLogger(__name__)


class UsedTablesCrawler(CrawlerBase[UsedTable]):

    def __init__(self, backend: SqlBackend, schema: str, table: str) -> None:
        """
        Initializes a DFSACrawler instance.

        Args:
            sql_backend (SqlBackend): The SQL Execution Backend abstraction (either REST API or Spark)
            schema: The schema name for the inventory persistence.
        """
        super().__init__(backend=backend, catalog="hive_metastore", schema=schema, table=table, klass=UsedTable)

    @classmethod
    def for_paths(cls, backend: SqlBackend, schema: str) -> UsedTablesCrawler:
        return UsedTablesCrawler(backend, schema, "used_tables_in_paths")

    @classmethod
    def for_queries(cls, backend: SqlBackend, schema: str) -> UsedTablesCrawler:
        return UsedTablesCrawler(backend, schema, "used_tables_in_queries")

    def dump_all(self, tables: Sequence[UsedTable]) -> None:
        """This crawler doesn't follow the pull model because the fetcher fetches data for 3 crawlers, not just one
        It's not **bad** because all records are pushed at once.
        Providing a multi-entity crawler is out-of-scope of this PR
        """
        try:
            # TODO until we historize data, we append all TableInfos
            self._update_snapshot(tables, mode="append")
        except DatabricksError as e:
            logger.error("Failed to store DFSAs", exc_info=e)

    def _try_fetch(self) -> Iterable[UsedTable]:
        sql = f"SELECT * FROM {escape_sql_identifier(self.full_name)}"
        for row in self._backend.fetch(sql):
            yield self._klass.from_dict(row.as_dict())

    def _crawl(self) -> Iterable[UsedTable]:
        return []
        # TODO raise NotImplementedError() once CrawlerBase supports empty snapshots



import dataclasses
import logging
from collections import defaultdict
from functools import partial

from databricks.labs.blueprint.parallel import Threads
from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.sdk import WorkspaceClient

from databricks.labs.ucx.hive_metastore import TablesCrawler, Mounts
from databricks.labs.ucx.hive_metastore.grants import MigrateGrants
from databricks.labs.ucx.hive_metastore.locations import Mount, ExternalLocations
from databricks.labs.ucx.hive_metastore.mapping import (
    Rule,
    TableMapping,
    TableToMigrate,
)
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationStatusRefresher
from databricks.labs.ucx.hive_metastore.tables import (
    MigrationCount,
    Table,
    What,
    HiveSerdeType,
)
from databricks.labs.ucx.hive_metastore.view_migrate import (
    ViewsMigrationSequencer,
    ViewToMigrate,
)
from databricks.sdk.errors.platform import DatabricksError

logger = logging.getLogger(__name__)


class TablesMigrator:
    def __init__(
        self,
        table_crawler: TablesCrawler,
        ws: WorkspaceClient,
        backend: SqlBackend,
        table_mapping: TableMapping,
        migration_status_refresher: TableMigrationStatusRefresher,
        migrate_grants: MigrateGrants,
    ):
        self._tc = table_crawler
        self._backend = backend
        self._ws = ws
        self._tm = table_mapping
        self._migration_status_refresher = migration_status_refresher
        self._seen_tables: dict[str, str] = {}
        self._migrate_grants = migrate_grants

    def get_remaining_tables(self) -> list[Table]:
        self.index(force_refresh=True)
        table_rows = []
        for crawled_table in self._tc.snapshot():
            if not self._is_migrated(crawled_table.database, crawled_table.name):
                table_rows.append(crawled_table)
                logger.warning(f"remained-hive-metastore-table: {crawled_table.key}")
        return table_rows

    def index(self, *, force_refresh: bool = False):
        return self._migration_status_refresher.index(force_refresh=force_refresh)

    def migrate_tables(
        self,
        what: What,
        mounts_crawler: Mounts | None = None,
        hiveserde_in_place_migrate: bool = False,
    ):
        if what in [What.DB_DATASET, What.UNKNOWN]:
            logger.error(f"Can't migrate tables with type {what.name}")
            return None
        self._init_seen_tables()
        # mounts will be used to replace the mnt based table location in the DDL for hiveserde table in-place migration
        mounts: list[Mount] = []
        if mounts_crawler:
            mounts = list(mounts_crawler.snapshot())
        if what == What.VIEW:
            return self._migrate_views()
        return self._migrate_tables(what, mounts, hiveserde_in_place_migrate)

    def _migrate_tables(self, what: What, mounts: list[Mount], hiveserde_in_place_migrate: bool = False):
        tables_to_migrate = self._tm.get_tables_to_migrate(self._tc)
        tables_in_scope = filter(lambda t: t.src.what == what, tables_to_migrate)
        tasks = []
        for table in tables_in_scope:
            tasks.append(partial(self._migrate_table, table, mounts, hiveserde_in_place_migrate))
        Threads.strict("migrate tables", tasks)
        if not tasks:
            logger.info(f"No tables found to migrate with type {what.name}")
        # the below is useful for testing
        return tasks

    def _migrate_views(self):
        tables_to_migrate = self._tm.get_tables_to_migrate(self._tc)
        all_tasks = []
        # Every batch of views to migrate needs an up-to-date table migration index
        # to determine if the dependencies have been migrated
        sequencer = ViewsMigrationSequencer(tables_to_migrate, migration_index=self.index(force_refresh=True))
        batches = sequencer.sequence_batches()
        for batch in batches:
            tasks = []
            for view in batch:
                tasks.append(partial(self._migrate_view, view))
            Threads.strict("migrate views", tasks)
            all_tasks.extend(tasks)
            self.index(force_refresh=True)
        return all_tasks

    def _migrate_table(self, src_table: TableToMigrate, mounts: list[Mount], hiveserde_in_place_migrate: bool = False):
        if self._table_already_migrated(src_table.rule.as_uc_table_key):
            logger.info(f"Table {src_table.src.key} already migrated to {src_table.rule.as_uc_table_key}")
            return True
        if src_table.src.what == What.DBFS_ROOT_DELTA:
            return self._migrate_dbfs_root_table(src_table.src, src_table.rule)
        if src_table.src.what == What.DBFS_ROOT_NON_DELTA:
            return self._migrate_table_create_ctas(src_table.src, src_table.rule, mounts)
        if src_table.src.what == What.EXTERNAL_SYNC:
            return self._migrate_external_table(src_table.src, src_table.rule)
        if src_table.src.what == What.TABLE_IN_MOUNT:
            return self._migrate_table_in_mount(src_table.src, src_table.rule)
        if src_table.src.what == What.EXTERNAL_HIVESERDE:
            # This hiveserde_in_place_migrate is used to determine if current hiveserde migration should use in-place migration or CTAS.
            # We will provide two workflows for hiveserde table migration:
            # 1. One will migrate all hiveserde tables using CTAS which we officially support.
            # 2. The other one will migrate certain types of hiveserde in place, which is technically working, but the user
            # need to accept the risk that the old files created by hiveserde may not be processed correctly by Spark
            # datasource in corner cases.
            # User will need to decide which workflow to runs first which will migrate the hiveserde tables and mark the
            # `upgraded_to` property and hence those tables will be skipped in the migration workflow runs later.
            if hiveserde_in_place_migrate:
                return self._migrate_external_table_hiveserde_in_place(src_table.src, src_table.rule, mounts)
            return self._migrate_table_create_ctas(src_table.src, src_table.rule, mounts)
        if src_table.src.what == What.EXTERNAL_NO_SYNC:
            # use CTAS if table cannot be upgraded using SYNC and table is not hiveserde table
            return self._migrate_table_create_ctas(src_table.src, src_table.rule, mounts)
        logger.info(f"Table {src_table.src.key} is not supported for migration")
        return True

    def _migrate_view(self, src_view: ViewToMigrate):
        if self._table_already_migrated(src_view.rule.as_uc_table_key):
            logger.info(f"View {src_view.src.key} already migrated to {src_view.rule.as_uc_table_key}")
            return True
        if self._view_can_be_migrated(src_view):
            return self._migrate_view_table(src_view)
        logger.info(f"View {src_view.src.key} is not supported for migration")
        return True

    def _view_can_be_migrated(self, view: ViewToMigrate):
        # dependencies have already been computed, therefore an empty dict is good enough
        for table in view.dependencies:
            if not self.index().get(table.schema, table.name):
                logger.info(f"View {view.src.key} cannot be migrated because {table.key} is not migrated yet")
                return False
        return True

    def _migrate_view_table(self, src_view: ViewToMigrate):
        view_migrate_sql = self._sql_migrate_view(src_view)
        logger.debug(f"Migrating view {src_view.src.key} to using SQL query: {view_migrate_sql}")
        try:
            self._backend.execute(view_migrate_sql)
            self._backend.execute(self._sql_alter_to(src_view.src, src_view.rule.as_uc_table_key))
            self._backend.execute(
                self._sql_alter_from(src_view.src, src_view.rule.as_uc_table_key, self._ws.get_workspace_id())
            )
        except DatabricksError as e:
            logger.warning(f"Failed to migrate view {src_view.src.key} to {src_view.rule.as_uc_table_key}: {e}")
            return False
        return self._migrate_grants.apply(src_view.src, src_view.rule.as_uc_table_key)

    def _sql_migrate_view(self, src_view: ViewToMigrate) -> str:
        # We have to fetch create statement this way because of columns in:
        # CREATE VIEW x.y (col1, col2) AS SELECT * FROM w.t
        create_statement = self._backend.fetch(f"SHOW CREATE TABLE {src_view.src.safe_sql_key}")
        src_view.src.view_text = next(iter(create_statement))["createtab_stmt"]
        # this does not require the index to be refreshed because the dependencies have already been validated
        return src_view.sql_migrate_view(self.index())

    def _migrate_external_table(self, src_table: Table, rule: Rule):
        if src_table.object_type == "MANAGED":
            logger.warning(
                f"failed-to-migrate: Detected MANAGED table {src_table.name} on external storage, not currently "
                f"supported by UCX"
            )
            return False
        target_table_key = rule.as_uc_table_key
        table_migrate_sql = src_table.sql_migrate_external(target_table_key)
        logger.debug(f"Migrating external table {src_table.key} to using SQL query: {table_migrate_sql}")
        # have to wrap the fetch result with iter() for now, because StatementExecutionBackend returns iterator but RuntimeBackend returns list.
        sync_result = next(iter(self._backend.fetch(table_migrate_sql)))
        if sync_result.status_code != "SUCCESS":
            logger.warning(
                f"failed-to-migrate: SYNC command failed to migrate table {src_table.key} to {target_table_key}. "
                f"Status code: {sync_result.status_code}. Description: {sync_result.description}"
            )
            return False
        self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
        return self._migrate_grants.apply(src_table, rule.as_uc_table_key)

    def _migrate_external_table_hiveserde_in_place(self, src_table: Table, rule: Rule, mounts: list[Mount]):
        # verify hive serde type
        hiveserde_type = src_table.hiveserde_type(self._backend)
        if hiveserde_type in [
            HiveSerdeType.NOT_HIVESERDE,
            HiveSerdeType.OTHER_HIVESERDE,
            HiveSerdeType.INVALID_HIVESERDE_INFO,
        ]:
            logger.warning(f"{src_table.key} table can only be migrated using CTAS.")
            return False

        # if the src table location is using mount, resolve the mount location so it will be used in the updated DDL
        dst_table_location = None
        if mounts and src_table.is_dbfs_mnt:
            dst_table_location = ExternalLocations.resolve_mount(src_table.location, mounts)

        table_migrate_sql = src_table.sql_migrate_external_hiveserde_in_place(
            rule.catalog_name, rule.dst_schema, rule.dst_table, self._backend, hiveserde_type, dst_table_location
        )
        if not table_migrate_sql:
            logger.error(
                f"Failed to generate in-place migration DDL for {src_table.key}, skip the in-place migration. It can be migrated in CTAS workflow"
            )
            return False

        logger.debug(
            f"Migrating external table {src_table.key} to {rule.as_uc_table_key} using SQL query: {table_migrate_sql}"
        )
        try:
            self._backend.execute(table_migrate_sql)
            self._backend.execute(self._sql_alter_to(src_table, rule.as_uc_table_key))
            self._backend.execute(self._sql_add_migrated_comment(src_table, rule.as_uc_table_key))
            self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
        except DatabricksError as e:
            logger.warning(f"failed-to-migrate: Failed to migrate table {src_table.key} to {rule.as_uc_table_key}: {e}")
            return False
        return self._migrate_grants.apply(src_table, rule.as_uc_table_key)

    def _migrate_dbfs_root_table(self, src_table: Table, rule: Rule):
        target_table_key = rule.as_uc_table_key
        table_migrate_sql = src_table.sql_migrate_dbfs(target_table_key)
        logger.debug(
            f"Migrating managed table {src_table.key} to {rule.as_uc_table_key} using SQL query: {table_migrate_sql}"
        )
        try:
            self._backend.execute(table_migrate_sql)
            self._backend.execute(self._sql_alter_to(src_table, rule.as_uc_table_key))
            self._backend.execute(self._sql_add_migrated_comment(src_table, rule.as_uc_table_key))
            self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
        except DatabricksError as e:
            logger.warning(f"failed-to-migrate: Failed to migrate table {src_table.key} to {rule.as_uc_table_key}: {e}")
            return False
        return self._migrate_grants.apply(src_table, rule.as_uc_table_key)

    def _migrate_table_create_ctas(self, src_table: Table, rule: Rule, mounts: list[Mount]):
        if src_table.what not in [What.EXTERNAL_NO_SYNC, What.EXTERNAL_HIVESERDE]:
            table_migrate_sql = src_table.sql_migrate_ctas_managed(rule.as_uc_table_key)
        elif not src_table.location:
            table_migrate_sql = src_table.sql_migrate_ctas_managed(rule.as_uc_table_key)
        else:
            # if external table and src tabel location is not missing, migrate to external UC table
            dst_table_location = src_table.location + "_ctas_migrated"
            if mounts and src_table.is_dbfs_mnt:
                dst_table_location = ExternalLocations.resolve_mount(src_table.location, mounts) + "_ctas_migrated"
            table_migrate_sql = src_table.sql_migrate_ctas_external(rule.as_uc_table_key, dst_table_location)
        logger.debug(f"Migrating table {src_table.key} to {rule.as_uc_table_key} using SQL query: {table_migrate_sql}")
        try:
            self._backend.execute(table_migrate_sql)
            self._backend.execute(self._sql_alter_to(src_table, rule.as_uc_table_key))
            self._backend.execute(self._sql_add_migrated_comment(src_table, rule.as_uc_table_key))
            self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
        except DatabricksError as e:
            logger.warning(f"failed-to-migrate: Failed to migrate table {src_table.key} to {rule.as_uc_table_key}: {e}")
            return False
        return self._migrate_grants.apply(src_table, rule.as_uc_table_key)

    def _migrate_table_in_mount(self, src_table: Table, rule: Rule):
        target_table_key = rule.as_uc_table_key
        try:
            table_schema = self._backend.fetch(f"DESCRIBE TABLE delta.`{src_table.location}`;")
            table_migrate_sql = src_table.sql_migrate_table_in_mount(target_table_key, table_schema)
            logger.info(
                f"Migrating table in mount {src_table.location} to UC table {rule.as_uc_table_key} using SQL query: {table_migrate_sql}"
            )
            self._backend.execute(table_migrate_sql)
            self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
        except DatabricksError as e:
            logger.warning(f"failed-to-migrate: Failed to migrate table {src_table.key} to {rule.as_uc_table_key}: {e}")
            return False
        return self._migrate_grants.apply(src_table, rule.as_uc_table_key)

    def _table_already_migrated(self, target) -> bool:
        return target in self._seen_tables

    def _get_tables_to_revert(self, schema: str | None = None, table: str | None = None) -> list[Table]:
        schema = schema.lower() if schema else None
        table = table.lower() if table else None
        reverse_seen = {v: k for (k, v) in self._seen_tables.items()}
        migrated_tables = []
        if table and not schema:
            logger.error("Cannot accept 'Table' parameter without 'Schema' parameter")
        for cur_table in self._tc.snapshot():
            if schema and cur_table.database != schema:
                continue
            if table and cur_table.name != table:
                continue
            if cur_table.key in reverse_seen:
                updated_table = dataclasses.replace(cur_table, upgraded_to=reverse_seen[cur_table.key])
                migrated_tables.append(updated_table)
        return migrated_tables

    def revert_migrated_tables(
        self, schema: str | None = None, table: str | None = None, *, delete_managed: bool = False
    ):
        self._init_seen_tables()
        migrated_tables = self._get_tables_to_revert(schema=schema, table=table)
        # reverses the _seen_tables dictionary to key by the source table
        reverse_seen = {v: k for (k, v) in self._seen_tables.items()}
        tasks = []
        for migrated_table in migrated_tables:
            if migrated_table.kind == "VIEW" or migrated_table.object_type == "EXTERNAL" or delete_managed:
                tasks.append(partial(self._revert_migrated_table, migrated_table, reverse_seen[migrated_table.key]))
                continue
            logger.info(
                f"Skipping {migrated_table.object_type} Table {migrated_table.database}.{migrated_table.name} "
                f"upgraded_to {migrated_table.upgraded_to}"
            )
        Threads.strict("revert migrated tables", tasks)

    def _revert_migrated_table(self, table: Table, target_table_key: str):
        logger.info(
            f"Reverting {table.object_type} table {table.database}.{table.name} upgraded_to {table.upgraded_to}"
        )
        try:
            self._backend.execute(table.sql_unset_upgraded_to())
            self._backend.execute(f"DROP {table.kind} IF EXISTS {escape_sql_identifier(target_table_key)}")
        except DatabricksError as e:
            logger.warning(f"Failed to revert table {table.key}: {e}")

    def _get_revert_count(self, schema: str | None = None, table: str | None = None) -> list[MigrationCount]:
        self._init_seen_tables()
        migrated_tables = self._get_tables_to_revert(schema=schema, table=table)
        table_by_database = defaultdict(list)
        for cur_table in migrated_tables:
            table_by_database[cur_table.database].append(cur_table)

        migration_list = []
        for cur_database, tables in table_by_database.items():
            what_count: dict[What, int] = {}
            for current_table in tables:
                if current_table.upgraded_to is not None:
                    count = what_count.get(current_table.what, 0)
                    what_count[current_table.what] = count + 1
            migration_list.append(MigrationCount(database=cur_database, what_count=what_count))
        return migration_list

    def is_migrated(self, schema: str, table: str) -> bool:
        return self._migration_status_refresher.is_migrated(schema, table)

    def print_revert_report(
        self,
        *,
        schema: str | None = None,
        table: str | None = None,
        delete_managed: bool,
    ) -> bool | None:
        migrated_count = self._get_revert_count(schema=schema, table=table)
        if not migrated_count:
            logger.info("No migrated tables were found.")
            return False
        print("The following is the count of migrated tables and views found in scope:")
        table_header = "Database            |"
        headers = 1
        for what in list(What):
            headers = max(headers, len(what.name.split("_")))
            table_header += f" {what.name.split('_')[0]:<10} |"
        print(table_header)
        # Split the header so _ separated what names are splitted into multiple lines
        for header in range(1, headers):
            table_sub_header = "                    |"
            for what in list(What):
                if len(what.name.split("_")) - 1 < header:
                    table_sub_header += f"{' ' * 12}|"
                    continue
                table_sub_header += f" {what.name.split('_')[header]:<10} |"
            print(table_sub_header)
        separator = "=" * (22 + 13 * len(What))
        print(separator)
        for count in migrated_count:
            table_row = f"{count.database:<20}|"
            for what in list(What):
                table_row += f" {count.what_count.get(what, 0):10} |"
            print(table_row)
        print(separator)
        print("The following actions will be performed")
        print("- Migrated External Tables and Views (targets) will be deleted")
        if delete_managed:
            print("- Migrated DBFS Root Tables will be deleted")
        else:
            print("- Migrated DBFS Root Tables will be left intact")
            print("To revert and delete Migrated Tables, add --delete_managed true flag to the command")
        return True

    def _init_seen_tables(self):
        self._seen_tables = self._migration_status_refresher.get_seen_tables()

    def _sql_alter_to(self, table: Table, target_table_key: str):
        return f"ALTER {table.kind} {escape_sql_identifier(table.key)} SET TBLPROPERTIES ('upgraded_to' = '{target_table_key}');"

    def _sql_add_migrated_comment(self, table: Table, target_table_key: str):
        """Docs: https://docs.databricks.com/en/data-governance/unity-catalog/migrate.html#add-comments-to-indicate-that-a-hive-table-has-been-migrated"""
        return f"COMMENT ON {table.kind} {escape_sql_identifier(table.key)} IS 'This {table.kind.lower()} is deprecated. Please use `{target_table_key}` instead of `{table.key}`.';"

    def _sql_alter_from(self, table: Table, target_table_key: str, ws_id: int):
        source = table.location if table.is_table_in_mount else table.key
        return (
            f"ALTER {table.kind} {escape_sql_identifier(target_table_key)} SET TBLPROPERTIES "
            f"('upgraded_from' = '{source}'"
            f" , '{table.UPGRADED_FROM_WS_PARAM}' = '{ws_id}');"
        )

    def _is_migrated(self, schema: str, table: str) -> bool:
        index = self._migration_status_refresher.index()
        return index.is_migrated(schema, table)



import collections
import logging
from dataclasses import replace
import fnmatch
from pathlib import PurePath

from databricks.labs.blueprint.tui import Prompts
from databricks.labs.lsql.backends import SqlBackend
from databricks.labs.ucx.hive_metastore.grants import PrincipalACL, Grant, GrantsCrawler
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound
from databricks.sdk.service.catalog import SchemaInfo
from databricks.sdk.errors.platform import BadRequest

from databricks.labs.ucx.hive_metastore.mapping import TableMapping

logger = logging.getLogger(__name__)


class CatalogSchema:

    def __init__(
        self,
        ws: WorkspaceClient,
        table_mapping: TableMapping,
        principal_grants: PrincipalACL,
        sql_backend: SqlBackend,
        grants_crawler: GrantsCrawler,
        ucx_catalog: str,
    ):
        self._ws = ws
        self._table_mapping = table_mapping
        self._external_locations = self._ws.external_locations.list()
        self._principal_grants = principal_grants
        self._backend = sql_backend
        self._hive_grants_crawler = grants_crawler
        self._ucx_catalog = ucx_catalog

    def create_ucx_catalog(self, prompts: Prompts, *, properties: dict[str, str] | None = None) -> None:
        """Create the UCX catalog.

        Args:
            prompts : Prompts
                The prompts object to use for interactive input.
            properties : (dict[str, str] | None), default None
                The properties to pass to the catalog. If None, no properties are passed.
        """
        try:
            self._create_catalog_validate(self._ucx_catalog, prompts, properties=properties)
        except BadRequest as e:
            if "already exists" in str(e):
                logger.warning(f"Catalog '{self._ucx_catalog}' already exists. Skipping.")
                return
            raise

    def create_all_catalogs_schemas(self, prompts: Prompts) -> None:
        candidate_catalogs, candidate_schemas = self._get_missing_catalogs_schemas()
        for candidate_catalog in candidate_catalogs:
            try:
                self._create_catalog_validate(candidate_catalog, prompts, properties=None)
            except BadRequest as e:
                if "already exists" in str(e):
                    logger.warning(f"Catalog '{candidate_catalog}' already exists. Skipping.")
                    continue
        for candidate_catalog, schemas in candidate_schemas.items():
            for candidate_schema in schemas:
                try:
                    self._create_schema(candidate_catalog, candidate_schema)
                except BadRequest as e:
                    if "already exists" in str(e):
                        logger.warning(
                            f"Schema '{candidate_schema}' in catalog '{candidate_catalog}' already exists. Skipping."
                        )
                        continue
        self._apply_from_legacy_table_acls()
        self._update_principal_acl()

    def _apply_from_legacy_table_acls(self):
        grants = self._get_catalog_schema_hive_grants()
        for grant in grants:
            acl_migrate_sql = grant.uc_grant_sql()
            if acl_migrate_sql is None:
                logger.warning(f"Cannot identify UC grant for {grant.this_type_and_key()}. Skipping.")
                continue
            logger.debug(f"Migrating acls on {grant.this_type_and_key()} using SQL query: {acl_migrate_sql}")
            self._backend.execute(acl_migrate_sql)

    def _update_principal_acl(self):

        grants = self._get_catalog_schema_principal_acl_grants()
        for grant in grants:
            acl_migrate_sql = grant.uc_grant_sql()
            if acl_migrate_sql is None:
                logger.warning(f"Cannot identify UC grant for {grant.this_type_and_key()}. Skipping.")
                continue
            logger.debug(f"Migrating acls on {grant.this_type_and_key()} using SQL query: {acl_migrate_sql}")
            self._backend.execute(acl_migrate_sql)

    def _get_catalog_schema_hive_grants(self) -> list[Grant]:
        src_dst_schema_mapping = self._get_database_source_target_mapping()
        hive_grants = self._hive_grants_crawler.snapshot()
        new_grants: list[Grant] = []
        for grant in hive_grants:
            if grant.this_type_and_key()[0] == "DATABASE" and grant.database:
                for schema in src_dst_schema_mapping[grant.database]:
                    new_grants.append(replace(grant, catalog=schema.catalog_name, database=schema.name))
        catalog_grants: set[Grant] = set()
        for grant in new_grants:
            catalog_grants.add(replace(grant, database=None))
        new_grants.extend(catalog_grants)
        return new_grants

    def _get_catalog_schema_principal_acl_grants(self) -> list[Grant]:
        src_trg_schema_mapping = self._get_database_source_target_mapping()
        grants = self._principal_grants.get_interactive_cluster_grants()
        # filter on grants to only get database level grants
        new_grants: list[Grant] = []
        for grant in grants:
            # For a database grant the table/view are not set, while the database is.
            if grant.table is None and grant.view is None:
                database = grant.database
                if database is not None:
                    new_grants.extend(
                        replace(grant, catalog=schema.catalog_name, database=schema.name)
                        for schema in src_trg_schema_mapping[database]
                    )
        catalog_grants: set[Grant] = set()
        for grant in new_grants:
            catalog_grants.add(replace(grant, database=None))
        new_grants.extend(catalog_grants)
        return new_grants

    def _get_database_source_target_mapping(self) -> dict[str, list[SchemaInfo]]:
        """Generate a dictionary of source database in hive_metastore and its
        mapping of target UC catalog and schema combinations from the table mappings."""
        src_trg_schema_mapping: dict[str, list[SchemaInfo]] = collections.defaultdict(list)
        table_mappings = self._table_mapping.load()
        for table_mapping in table_mappings:
            schema = SchemaInfo(catalog_name=table_mapping.catalog_name, name=table_mapping.dst_schema)
            if schema not in src_trg_schema_mapping[table_mapping.src_schema]:
                src_trg_schema_mapping[table_mapping.src_schema].append(schema)
        return src_trg_schema_mapping

    def _create_catalog_validate(self, catalog: str, prompts: Prompts, *, properties: dict[str, str] | None) -> None:
        logger.info(f"Validating UC catalog: {catalog}")
        attempts = 3
        while True:
            catalog_storage = prompts.question(
                f"Please provide storage location url for catalog: {catalog}", default="metastore"
            )
            if self._validate_location(catalog_storage):
                break
            attempts -= 1
            if attempts == 0:
                raise NotFound(f"Failed to validate location for {catalog} catalog")
        self._create_catalog(catalog, catalog_storage, properties=properties)

    def _list_existing(self) -> tuple[set[str], dict[str, set[str]]]:
        """generate a list of existing UC catalogs and schema."""
        logger.info("Listing existing UC catalogs and schemas")
        existing_catalogs: set[str] = set()
        for catalog_info in self._ws.catalogs.list():
            if catalog_info.name:
                existing_catalogs.add(catalog_info.name)

        existing_schemas: dict[str, set[str]] = {}  # catalog -> set[schema]
        for catalog in existing_catalogs:
            existing_schemas[catalog] = set()
            for schema in self._ws.schemas.list(catalog, max_results=0):
                if schema.name:
                    existing_schemas[catalog].add(schema.name)

        return existing_catalogs, existing_schemas

    def _list_target(self) -> tuple[set[str], dict[str, set[str]]]:
        """generate a list of catalogs and schema to be created from table mappings."""
        target_catalogs: set[str] = set()
        target_schemas: dict[str, set[str]] = {}  # catalog -> set[schema]
        table_mappings = self._table_mapping.load()
        for mappings in table_mappings:
            target_catalog = mappings.catalog_name
            target_schema = mappings.dst_schema
            target_catalogs.add(target_catalog)
            if target_catalog not in target_schemas:
                target_schemas[target_catalog] = {target_schema}
                continue
            target_schemas[target_catalog].add(target_schema)
        return target_catalogs, target_schemas

    def _get_missing_catalogs_schemas(self) -> tuple[set[str], dict[str, set[str]]]:
        """prepare a list of catalogs and schema to be created"""
        existing_catalogs, existing_schemas = self._list_existing()
        target_catalogs, target_schemas = self._list_target()

        logger.info("Preparing a list of UC catalogs and schema to be created")
        # filter out existing catalogs and schemas from target catalogs and schemas to be created.
        for existing_catalog in existing_catalogs:
            if existing_catalog in target_catalogs:
                target_catalogs.remove(existing_catalog)

        for catalog, schemas in existing_schemas.items():
            if catalog in target_schemas:
                target_schemas[catalog] = target_schemas[catalog] - schemas
        return target_catalogs, target_schemas

    def _validate_location(self, location: str):
        if location == "metastore":
            return True
        try:
            PurePath(location)
        except ValueError:
            logger.error(f"Invalid location path {location}")
            return False
        for external_location in self._external_locations:
            if location == external_location.url:
                return True
            if external_location.url is not None and fnmatch.fnmatch(location, external_location.url + '*'):
                return True
        return False

    def _create_catalog(self, catalog: str, catalog_storage: str, *, properties: dict[str, str] | None) -> None:
        logger.info(f"Creating UC catalog: {catalog}")
        if catalog_storage == "metastore":
            self._ws.catalogs.create(catalog, comment="Created by UCX", properties=properties)
        else:
            self._ws.catalogs.create(
                catalog,
                storage_root=catalog_storage,
                comment="Created by UCX",
                properties=properties,
            )

    def _create_schema(self, catalog, schema):
        logger.info(f"Creating UC schema: {schema} in catalog: {catalog}")
        self._ws.schemas.create(schema, catalog, comment="Created by UCX")



from databricks.labs.ucx.contexts.workflow_task import RuntimeContext
from databricks.labs.ucx.framework.tasks import Workflow, job_task


class MigrationRecon(Workflow):
    def __init__(self):
        super().__init__('migrate-data-reconciliation')

    @job_task(job_cluster="table_migration")
    def recon_migration_result(self, ctx: RuntimeContext):
        """This workflow validate post-migration datasets against their pre-migration counterparts. This includes all
        tables, by comparing their schema, row counts and row comparison
        """
        # need to delete the existing content of recon results table, so that snapshot will re-populate it
        ctx.migration_recon.reset()
        ctx.migration_recon.snapshot()



import dataclasses
import functools
import logging
import os
import re
import time
import webbrowser
from collections.abc import Callable, Iterable
from datetime import timedelta
from functools import cached_property
from pathlib import Path
from typing import Any

from requests.exceptions import ConnectionError as RequestsConnectionError

import databricks.sdk.errors
from databricks.labs.blueprint.entrypoint import get_logger, is_in_debug
from databricks.labs.blueprint.installation import Installation, SerdeError
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.parallel import ManyError, Threads
from databricks.labs.blueprint.tui import Prompts
from databricks.labs.blueprint.upgrades import Upgrades
from databricks.labs.blueprint.wheels import (
    ProductInfo,
    Version,
    find_project_root,
)
from databricks.labs.lsql.backends import SqlBackend, StatementExecutionBackend
from databricks.labs.lsql.dashboards import DashboardMetadata, Dashboards
from databricks.labs.lsql.deployment import SchemaDeployer
from databricks.sdk import WorkspaceClient, AccountClient
from databricks.sdk.useragent import with_extra
from databricks.sdk.errors import (
    AlreadyExists,
    BadRequest,
    DeadlineExceeded,
    InternalError,
    InvalidParameterValue,
    NotFound,
    PermissionDenied,
    ResourceAlreadyExists,
    ResourceDoesNotExist,
    Unauthenticated,
    OperationFailed,
)
from databricks.sdk.retries import retried
from databricks.sdk.service.dashboards import LifecycleState
from databricks.sdk.service.provisioning import Workspace
from databricks.sdk.service.sql import (
    CreateWarehouseRequestWarehouseType,
    EndpointInfoWarehouseType,
    SpotInstancePolicy,
)

from databricks.labs.ucx.__about__ import __version__
from databricks.labs.ucx.assessment.azure import AzureServicePrincipalInfo
from databricks.labs.ucx.assessment.clusters import ClusterInfo, PolicyInfo
from databricks.labs.ucx.assessment.init_scripts import GlobalInitScriptInfo
from databricks.labs.ucx.assessment.jobs import JobInfo, SubmitRunInfo
from databricks.labs.ucx.assessment.pipelines import PipelineInfo
from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.ucx.contexts.account_cli import AccountContext
from databricks.labs.ucx.contexts.workspace_cli import WorkspaceContext
from databricks.labs.ucx.framework.tasks import Task
from databricks.labs.ucx.hive_metastore.grants import Grant
from databricks.labs.ucx.hive_metastore.locations import ExternalLocation, Mount
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationStatus
from databricks.labs.ucx.hive_metastore.table_size import TableSize
from databricks.labs.ucx.hive_metastore.tables import Table, TableError
from databricks.labs.ucx.hive_metastore.udfs import Udf
from databricks.labs.ucx.installer.hms_lineage import HiveMetastoreLineageEnabler
from databricks.labs.ucx.installer.logs import LogRecord
from databricks.labs.ucx.installer.mixins import InstallationMixin
from databricks.labs.ucx.installer.policy import ClusterPolicyInstaller
from databricks.labs.ucx.installer.workflows import WorkflowsDeployment
from databricks.labs.ucx.recon.migration_recon import ReconResult
from databricks.labs.ucx.runtime import Workflows
from databricks.labs.ucx.source_code.base import UsedTable
from databricks.labs.ucx.source_code.directfs_access import DirectFsAccess
from databricks.labs.ucx.source_code.jobs import JobProblem
from databricks.labs.ucx.source_code.queries import QueryProblem
from databricks.labs.ucx.workspace_access.base import Permissions
from databricks.labs.ucx.workspace_access.generic import WorkspaceObjectInfo
from databricks.labs.ucx.workspace_access.groups import ConfigureGroups, MigratedGroup

TAG_STEP = "step"
WAREHOUSE_PREFIX = "Unity Catalog Migration"
NUM_USER_ATTEMPTS = 10  # number of attempts user gets at answering a question

logger = logging.getLogger(__name__)


def deploy_schema(sql_backend: SqlBackend, inventory_schema: str):
    # we need to import it like this because we expect a module instance
    from databricks.labs import ucx  # pylint: disable=import-outside-toplevel

    logger.info("Creating ucx schemas...")
    deployer = SchemaDeployer(sql_backend, inventory_schema, ucx)
    deployer.deploy_schema()
    table = functools.partial(deployer.deploy_table)
    Threads.strict(
        "deploy tables",
        [
            functools.partial(table, "azure_service_principals", AzureServicePrincipalInfo),
            functools.partial(table, "clusters", ClusterInfo),
            functools.partial(table, "global_init_scripts", GlobalInitScriptInfo),
            functools.partial(table, "jobs", JobInfo),
            functools.partial(table, "pipelines", PipelineInfo),
            functools.partial(table, "external_locations", ExternalLocation),
            functools.partial(table, "mounts", Mount),
            functools.partial(table, "grants", Grant),
            functools.partial(table, "groups", MigratedGroup),
            functools.partial(table, "tables", Table),
            functools.partial(table, "table_size", TableSize),
            functools.partial(table, "table_failures", TableError),
            functools.partial(table, "workspace_objects", WorkspaceObjectInfo),
            functools.partial(table, "permissions", Permissions),
            functools.partial(table, "submit_runs", SubmitRunInfo),
            functools.partial(table, "policies", PolicyInfo),
            functools.partial(table, "migration_status", TableMigrationStatus),
            functools.partial(table, "workflow_problems", JobProblem),
            functools.partial(table, "query_problems", QueryProblem),
            functools.partial(table, "udfs", Udf),
            functools.partial(table, "logs", LogRecord),
            functools.partial(table, "recon_results", ReconResult),
            functools.partial(table, "directfs_in_paths", DirectFsAccess),
            functools.partial(table, "directfs_in_queries", DirectFsAccess),
            functools.partial(table, "used_tables_in_paths", UsedTable),
            functools.partial(table, "used_tables_in_queries", UsedTable),
        ],
    )
    deployer.deploy_view("grant_detail", "queries/views/grant_detail.sql")
    deployer.deploy_view("objects", "queries/views/objects.sql")
    deployer.deploy_view("table_estimates", "queries/views/table_estimates.sql")
    deployer.deploy_view("misc_patterns", "queries/views/misc_patterns.sql")
    deployer.deploy_view("code_patterns", "queries/views/code_patterns.sql")
    deployer.deploy_view("reconciliation_results", "queries/views/reconciliation_results.sql")
    deployer.deploy_view("directfs", "queries/views/directfs.sql")
    deployer.deploy_view("used_tables", "queries/views/used_tables.sql")


def extract_major_minor(version_string):
    match = re.search(r'(\d+\.\d+)', version_string)
    if match:
        return match.group(1)
    return None


class WorkspaceInstaller(WorkspaceContext):
    def __init__(
        self,
        ws: WorkspaceClient,
        environ: dict[str, str] | None = None,
        tasks: list[Task] | None = None,
    ):
        super().__init__(ws)
        if not environ:
            environ = dict(os.environ.items())
        self._force_install = environ.get("UCX_FORCE_INSTALL")
        if "DATABRICKS_RUNTIME_VERSION" in environ:
            msg = "WorkspaceInstaller is not supposed to be executed in Databricks Runtime"
            raise SystemExit(msg)

        self._is_account_install = self._force_install == "account"
        self._tasks = tasks if tasks else Workflows.all().tasks()

    @cached_property
    def upgrades(self) -> Upgrades:
        return Upgrades(self.product_info, self.installation)

    @cached_property
    def policy_installer(self) -> ClusterPolicyInstaller:
        return ClusterPolicyInstaller(self.installation, self.workspace_client, self.prompts)

    @cached_property
    def installation(self) -> Installation:
        try:
            return self.product_info.current_installation(self.workspace_client)
        except NotFound:
            if self._force_install == "user":
                return Installation.assume_user_home(self.workspace_client, self.product_info.product_name())
            return Installation.assume_global(self.workspace_client, self.product_info.product_name())

    def run(
        self,
        default_config: WorkspaceConfig | None = None,
        config: WorkspaceConfig | None = None,
    ) -> WorkspaceConfig:
        logger.info(f"Installing UCX v{self.product_info.version()}")
        try:
            if config is None:
                config = self.configure(default_config)
            if self._is_testing():
                return config
            workflows_deployment = WorkflowsDeployment(
                config,
                self.installation,
                self.install_state,
                self.workspace_client,
                self.wheels,
                self.product_info,
                self._tasks,
            )
            workspace_installation = WorkspaceInstallation(
                config,
                self.installation,
                self.install_state,
                self.sql_backend,
                self.workspace_client,
                workflows_deployment,
                self.prompts,
                self.product_info,
            )
            workspace_installation.run()
        except ManyError as err:
            if len(err.errs) == 1:
                raise err.errs[0] from None
            raise err
        except TimeoutError as err:
            if isinstance(err.__cause__, RequestsConnectionError):
                logger.warning(
                    f"Cannot connect with {self.workspace_client.config.host} see "
                    f"https://github.com/databrickslabs/ucx#network-connectivity-issues for help: {err}"
                )
            raise err
        return config

    def _is_testing(self):
        return self.product_info.product_name() != "ucx"

    def _prompt_for_new_installation(self) -> WorkspaceConfig:
        logger.info("Please answer a couple of questions to configure Unity Catalog migration")
        default_database = "ucx"
        # if a workspace is configured to use external hive metastore, the majority of the time that metastore will be
        # shared with other workspaces. we need to add the suffix to ensure uniqueness of the inventory database
        if self.policy_installer.has_ext_hms():
            default_database = f"ucx_{self.workspace_client.get_workspace_id()}"
        inventory_database = self.prompts.question(
            "Inventory Database stored in hive_metastore", default=default_database, valid_regex=r"^\w+$"
        )
        ucx_catalog = self.prompts.question("Catalog to store UCX artifacts in", default="ucx", valid_regex=r"^\w+$")
        log_level = self.prompts.question("Log level", default="INFO").upper()
        num_threads = int(self.prompts.question("Number of threads", default="8", valid_number=True))
        configure_groups = ConfigureGroups(self.prompts)
        configure_groups.run()
        include_databases = self._select_databases()
        upload_dependencies = self.prompts.confirm(
            f"Does given workspace {self.workspace_client.get_workspace_id()} block Internet access?"
        )
        trigger_job = self.prompts.confirm("Do you want to trigger assessment job after installation?")
        recon_tolerance_percent = int(
            self.prompts.question("Reconciliation threshold, in percentage", default="5", valid_number=True)
        )
        return WorkspaceConfig(
            inventory_database=inventory_database,
            ucx_catalog=ucx_catalog,
            workspace_group_regex=configure_groups.workspace_group_regex,
            workspace_group_replace=configure_groups.workspace_group_replace,
            account_group_regex=configure_groups.account_group_regex,
            group_match_by_external_id=configure_groups.group_match_by_external_id,  # type: ignore[arg-type]
            include_group_names=configure_groups.include_group_names,
            renamed_group_prefix=configure_groups.renamed_group_prefix,
            log_level=log_level,
            num_threads=num_threads,
            include_databases=include_databases,
            trigger_job=trigger_job,
            recon_tolerance_percent=recon_tolerance_percent,
            upload_dependencies=upload_dependencies,
        )

    def _compare_remote_local_versions(self):
        try:
            local_version = self.product_info.released_version()
            remote_version = self.installation.load(Version).version
            if extract_major_minor(remote_version) == extract_major_minor(local_version):
                logger.info(f"UCX v{self.product_info.version()} is already installed on this workspace")
                msg = "Do you want to update the existing installation?"
                if not self.prompts.confirm(msg):
                    raise RuntimeWarning(
                        "UCX workspace remote and local install versions are same and no override is requested. Exiting..."
                    )
        except NotFound as err:
            logger.warning(f"UCX workspace remote version not found: {err}")

    def _confirm_force_install(self) -> bool:
        if not self._force_install:
            return False
        msg = "[ADVANCED] UCX is already installed on this workspace. Do you want to create a new installation?"
        if not self.prompts.confirm(msg):
            raise RuntimeWarning("UCX is already installed, but no confirmation")
        if not self.installation.is_global() and self._force_install == "global":
            # TODO:
            # Logic for forced global over user install
            # Migration logic will go here
            # verify complains without full path, asks to raise NotImplementedError builtin
            raise databricks.sdk.errors.NotImplemented("Migration needed. Not implemented yet.")
        if self.installation.is_global() and self._force_install == "user":
            # Logic for forced user install over global install
            self.replace(
                installation=Installation.assume_user_home(self.workspace_client, self.product_info.product_name())
            )
            return True
        return False

    def configure(self, default_config: WorkspaceConfig | None = None) -> WorkspaceConfig:
        """Configure the workspaces

        Notes:
        1. Connection errors are not handled within this configure method.
        """
        try:
            config = self.installation.load(WorkspaceConfig)
            self._compare_remote_local_versions()
            if self._confirm_force_install():
                return self._configure_new_installation(default_config)
            self._apply_upgrades()
            return config
        except NotFound as err:
            logger.debug(f"Cannot find previous installation: {err}")
        except (PermissionDenied, SerdeError, ValueError, AttributeError):
            logger.warning(f"Existing installation at {self.installation.install_folder()} is corrupted. Skipping...")
        return self._configure_new_installation(default_config)

    def replace_config(self, **changes: Any) -> WorkspaceConfig | None:
        """
        Persist the list of workspaces where UCX is successfully installed in the config
        """
        try:
            config = self.installation.load(WorkspaceConfig)
            new_config = dataclasses.replace(config, **changes)
            self.installation.save(new_config)
        except (PermissionDenied, NotFound, ValueError):
            logger.warning(f"Failed to replace config for {self.workspace_client.config.host}")
            new_config = None
        return new_config

    def _apply_upgrades(self):
        try:
            self.upgrades.apply(self.workspace_client)
        except (InvalidParameterValue, NotFound) as err:
            logger.warning(f"Installed version is too old: {err}")

    def _configure_new_installation(self, default_config: WorkspaceConfig | None = None) -> WorkspaceConfig:
        if default_config is None:
            default_config = self._prompt_for_new_installation()
        HiveMetastoreLineageEnabler(self.workspace_client).apply(self.prompts, self._is_account_install)
        self._check_inventory_database_exists(default_config.inventory_database)
        warehouse_id = self.configure_warehouse()
        policy_id, instance_profile, spark_conf_dict, instance_pool_id = self.policy_installer.create(
            default_config.inventory_database
        )

        # Save configurable values for table migration cluster
        min_workers, max_workers, spark_conf_dict = self._config_table_migration(spark_conf_dict)

        config = dataclasses.replace(
            default_config,
            warehouse_id=warehouse_id,
            instance_profile=instance_profile,
            spark_conf=spark_conf_dict,
            min_workers=min_workers,
            max_workers=max_workers,
            policy_id=policy_id,
            instance_pool_id=instance_pool_id,
        )
        self.installation.save(config)
        if self._is_account_install:
            return config
        ws_file_url = self.installation.workspace_link(config.__file__)
        if self.prompts.confirm(f"Open config file in the browser and continue installing? {ws_file_url}"):
            webbrowser.open(ws_file_url)
        return config

    def _config_table_migration(self, spark_conf_dict) -> tuple[int, int, dict]:
        # parallelism will not be needed if backlog is fixed in https://databricks.atlassian.net/browse/ES-975874
        if self._is_account_install:
            return 1, 10, spark_conf_dict
        parallelism = self.prompts.question(
            "Parallelism for migrating dbfs root delta tables with deep clone", default="200", valid_number=True
        )
        if int(parallelism) > 200:
            spark_conf_dict.update({'spark.sql.sources.parallelPartitionDiscovery.parallelism': parallelism})
        # mix max workers for auto-scale migration job cluster
        min_workers = int(
            self.prompts.question(
                "Min workers for auto-scale job cluster for table migration", default="1", valid_number=True
            )
        )
        max_workers = int(
            self.prompts.question(
                "Max workers for auto-scale job cluster for table migration", default="10", valid_number=True
            )
        )
        return min_workers, max_workers, spark_conf_dict

    def _select_databases(self):
        selected_databases = self.prompts.question(
            "Comma-separated list of databases to migrate. If not specified, we'll use all "
            "databases in hive_metastore",
            default="<ALL>",
        )
        if selected_databases != "<ALL>":
            return [x.strip() for x in selected_databases.split(",")]
        return None

    def configure_warehouse(self) -> str:
        def warehouse_type(_):
            return _.warehouse_type.value if not _.enable_serverless_compute else "SERVERLESS"

        pro_warehouses = {"[Create new PRO SQL warehouse]": "create_new"} | {
            f"{_.name} ({_.id}, {warehouse_type(_)}, {_.state.value})": _.id
            for _ in self.workspace_client.warehouses.list()
            if _.warehouse_type == EndpointInfoWarehouseType.PRO
        }
        if self._is_account_install:
            warehouse_id = "create_new"
        else:
            warehouse_id = self.prompts.choice_from_dict(
                "Select PRO or SERVERLESS SQL warehouse to run assessment dashboards on", pro_warehouses
            )
        if warehouse_id == "create_new":
            new_warehouse = self.workspace_client.warehouses.create(
                name=f"{WAREHOUSE_PREFIX} {time.time_ns()}",
                spot_instance_policy=SpotInstancePolicy.COST_OPTIMIZED,
                warehouse_type=CreateWarehouseRequestWarehouseType.PRO,
                cluster_size="Small",
                max_num_clusters=1,
            )
            warehouse_id = new_warehouse.id
        return warehouse_id

    def _check_inventory_database_exists(self, inventory_database: str):
        logger.info("Fetching installations...")
        for installation in Installation.existing(self.workspace_client, self.product_info.product_name()):
            try:
                config = installation.load(WorkspaceConfig)
                if config.inventory_database == inventory_database:
                    raise AlreadyExists(
                        f"Inventory database '{inventory_database}' already exists in another installation"
                    )
            except (PermissionDenied, NotFound, SerdeError, ValueError, AttributeError):
                logger.warning(f"Existing installation at {installation.install_folder()} is corrupted. Skipping...")
                continue


class WorkspaceInstallation(InstallationMixin):
    def __init__(
        self,
        config: WorkspaceConfig,
        installation: Installation,
        install_state: InstallState,
        sql_backend: SqlBackend,
        ws: WorkspaceClient,
        workflows_installer: WorkflowsDeployment,
        prompts: Prompts,
        product_info: ProductInfo,
    ):
        self._config = config
        self._installation = installation
        self._install_state = install_state
        self._ws = ws
        self._sql_backend = sql_backend
        self._workflows_installer = workflows_installer
        self._prompts = prompts
        self._product_info = product_info
        environ = dict(os.environ.items())
        self._is_account_install = environ.get("UCX_FORCE_INSTALL") == "account"
        super().__init__(config, installation, ws)

    @classmethod
    def current(cls, ws: WorkspaceClient):
        # TODO: remove this method, it's no longer needed
        product_info = ProductInfo.from_class(WorkspaceConfig)
        installation = product_info.current_installation(ws)
        install_state = InstallState.from_installation(installation)
        config = installation.load(WorkspaceConfig)
        sql_backend = StatementExecutionBackend(ws, config.warehouse_id)
        wheels = product_info.wheels(ws)
        prompts = Prompts()
        tasks = Workflows.all().tasks()
        workflows_installer = WorkflowsDeployment(config, installation, install_state, ws, wheels, product_info, tasks)

        return cls(
            config,
            installation,
            install_state,
            sql_backend,
            ws,
            workflows_installer,
            prompts,
            product_info,
        )

    @property
    def config(self):
        return self._config

    @property
    def folder(self):
        return self._installation.install_folder()

    def run(self) -> bool:
        """Run workflow installation.

        Returns
            bool :
                True, installation finished. False installation did not finish.
        """
        logger.info(f"Installing UCX v{self._product_info.version()}")
        install_tasks = [self._create_database_and_dashboards, self._workflows_installer.create_jobs]
        Threads.strict("installing components", install_tasks)
        readme_url = self._installation.workspace_link("README")
        if not self._is_account_install and self._prompts.confirm(f"Open job overview in your browser? {readme_url}"):
            webbrowser.open(readme_url)
        logger.info(f"Installation completed successfully! Please refer to the {readme_url} for the next steps.")
        if self.config.trigger_job:
            logger.info("Triggering the assessment workflow")
            self._trigger_workflow("assessment")
        self._install_state.save()
        return True

    def _create_database_and_dashboards(self) -> None:
        self._create_database()  # Need the database before creating the dashboards
        Threads.strict("installing dashboards", list(self._get_create_dashboard_tasks()))

    # InternalError are retried for resilience on sporadic Databricks issues
    @retried(on=[InternalError], timeout=timedelta(minutes=2))
    def _create_database(self):
        try:
            deploy_schema(self._sql_backend, self._config.inventory_database)
        except Exception as err:
            if "UNRESOLVED_COLUMN.WITH_SUGGESTION" in str(err):
                msg = (
                    "The UCX version is not matching with the installed version."
                    "Kindly uninstall and reinstall UCX.\n"
                    "Please Follow the Below Command to uninstall and Install UCX\n"
                    "UCX Uninstall: databricks labs uninstall ucx.\n"
                    "UCX Install: databricks labs install ucx"
                )
                raise BadRequest(msg) from err
            if "Unable to load AWS credentials from any provider in the chain" in str(err):
                msg = (
                    "The UCX installation is configured to use external metastore. There is issue with the external metastore connectivity.\n"
                    "Please check the UCX installation instruction https://github.com/databrickslabs/ucx?tab=readme-ov-file#prerequisites"
                    "and re-run installation.\n"
                    "Please Follow the Below Command to uninstall and Install UCX\n"
                    "UCX Uninstall: databricks labs uninstall ucx.\n"
                    "UCX Install: databricks labs install ucx"
                )
                raise OperationFailed(msg) from err
            raise err

    def _get_create_dashboard_tasks(self) -> Iterable[Callable[[], None]]:
        """Get the tasks to create Lakeview dashboards from the SQL queries in the queries subfolders"""
        logger.info("Creating dashboards...")
        dashboard_folder_remote = f"{self._installation.install_folder()}/dashboards"
        try:
            self._ws.workspace.mkdirs(dashboard_folder_remote)
        except ResourceAlreadyExists:
            pass
        queries_folder = find_project_root(__file__) / "src/databricks/labs/ucx/queries"
        for step_folder in queries_folder.iterdir():
            if not step_folder.is_dir():
                continue
            logger.debug(f"Reading step folder {step_folder}...")
            for dashboard_folder in step_folder.iterdir():
                if not dashboard_folder.is_dir():
                    continue
                task = functools.partial(
                    self._create_dashboard,
                    dashboard_folder,
                    parent_path=dashboard_folder_remote,
                )
                yield task

    def _handle_existing_dashboard(self, dashboard_id: str, display_name: str, parent_path: str) -> str | None:
        """Handle an existing dashboard

        This method handles the following scenarios:
        - dashboard exists and needs to be updated
        - dashboard needs to be upgraded from Redash to Lakeview
        - dashboard is trashed and needs to be recreated
        - dashboard reference is invalid and the dashboard needs to be recreated

        Returns
            str | None :
                The dashboard id. If None, the dashboard will be recreated.
        """
        if "-" in dashboard_id:
            logger.info(f"Upgrading dashboard to Lakeview: {display_name} ({dashboard_id})")
            try:
                self._ws.dashboards.delete(dashboard_id=dashboard_id)
            except BadRequest:
                logger.warning(f"Cannot delete dashboard: {display_name} ({dashboard_id})")
            return None  # Recreate the dashboard if upgrading from Redash
        try:
            dashboard = self._ws.lakeview.get(dashboard_id)
            if dashboard.lifecycle_state is None:
                raise NotFound(f"Dashboard life cycle state: {display_name} ({dashboard_id})")
            if dashboard.lifecycle_state == LifecycleState.TRASHED:
                logger.info(f"Recreating trashed dashboard: {display_name} ({dashboard_id})")
                return None  # Recreate the dashboard if it is trashed (manually)
        except (NotFound, InvalidParameterValue):
            logger.info(f"Recovering invalid dashboard: {display_name} ({dashboard_id})")
            try:
                dashboard_path = f"{parent_path}/{display_name}.lvdash.json"
                self._ws.workspace.delete(dashboard_path)  # Cannot recreate dashboard if file still exists
                logger.debug(f"Deleted dangling dashboard {display_name} ({dashboard_id}): {dashboard_path}")
            except NotFound:
                pass
            return None  # Recreate the dashboard if it's reference is corrupted (manually)
        return dashboard_id  # Update the existing dashboard

    # InternalError and DeadlineExceeded are retried because of Lakeview internal issues
    # These issues have been reported to and are resolved by the Lakeview team
    # Keeping the retry for resilience
    @retried(on=[InternalError, DeadlineExceeded], timeout=timedelta(minutes=4))
    def _create_dashboard(self, folder: Path, *, parent_path: str) -> None:
        """Create a lakeview dashboard from the SQL queries in the folder"""
        logger.info(f"Creating dashboard in {folder}...")
        metadata = DashboardMetadata.from_path(folder).replace_database(
            database=f"hive_metastore.{self._config.inventory_database}",
            database_to_replace="inventory",
        )
        metadata.display_name = f"{self._name('UCX ')} {folder.parent.stem.title()} ({folder.stem.title()})"
        reference = f"{folder.parent.stem}_{folder.stem}".lower()
        dashboard_id = self._install_state.dashboards.get(reference)
        if dashboard_id is not None:
            dashboard_id = self._handle_existing_dashboard(dashboard_id, metadata.display_name, parent_path)
        dashboard = Dashboards(self._ws).create_dashboard(
            metadata,
            dashboard_id=dashboard_id,
            parent_path=parent_path,
            warehouse_id=self._warehouse_id,
            publish=True,
        )
        assert dashboard.dashboard_id is not None
        self._install_state.dashboards[reference] = dashboard.dashboard_id

    def uninstall(self):
        if self._prompts and not self._prompts.confirm(
            "Do you want to uninstall ucx from the workspace too, this would "
            "remove ucx project folder, dashboards, queries and jobs"
        ):
            return
        # TODO: this is incorrect, fetch the remote version (that appeared only in Feb 2024)
        logger.info(f"Deleting UCX v{self._product_info.version()} from {self._ws.config.host}")
        try:
            self._installation.files()
        except NotFound:
            logger.error(f"Check if {self._installation.install_folder()} is present")
            return
        self._check_and_fix_if_warehouse_does_not_exists()
        self._remove_database()
        self._remove_jobs()
        self._remove_warehouse()
        self._remove_policies()
        self._remove_secret_scope()
        self._installation.remove()
        logger.info("UnInstalling UCX complete")

    def _remove_database(self):
        if self._prompts and not self._prompts.confirm(
            f"Do you want to delete the inventory database {self._config.inventory_database} too?"
        ):
            return
        logger.info(f"Deleting inventory database {self._config.inventory_database}")
        deployer = SchemaDeployer(self._sql_backend, self._config.inventory_database, Any)
        deployer.delete_schema()

    def _remove_policies(self):
        logger.info("Deleting cluster policy")
        try:
            self._ws.cluster_policies.delete(policy_id=self.config.policy_id)
        except NotFound:
            logger.error("UCX Policy already deleted")

    def _remove_secret_scope(self):
        logger.info("Deleting secret scope")
        try:
            if self.config.uber_spn_id is not None:
                self._ws.secrets.delete_scope(self.config.inventory_database)
        except NotFound:
            logger.error("Secret scope already deleted")

    def _remove_jobs(self):
        self._workflows_installer.remove_jobs()

    def _remove_warehouse(self):
        try:
            warehouse_name = self._ws.warehouses.get(self._config.warehouse_id).name
            if warehouse_name.startswith(WAREHOUSE_PREFIX):
                logger.info(f"Deleting {warehouse_name}.")
                self._ws.warehouses.delete(id=self._config.warehouse_id)
        except InvalidParameterValue:
            logger.error("Error accessing warehouse details")

    def _trigger_workflow(self, step: str):
        job_id = int(self._install_state.jobs[step])
        job_url = f"{self._ws.config.host}#job/{job_id}"
        logger.debug(f"triggering {step} job: {self._ws.config.host}#job/{job_id}")
        self._ws.jobs.run_now(job_id)
        if self._prompts.confirm(f"Open {step} Job url that just triggered ? {job_url}"):
            webbrowser.open(job_url)

    def _check_and_fix_if_warehouse_does_not_exists(self):
        try:
            self._ws.warehouses.get(self._config.warehouse_id)
        except ResourceDoesNotExist:
            logger.critical(f"warehouse with id {self._config.warehouse_id} does not exists anymore")
            installer = WorkspaceInstaller(self._ws).replace(product_info=self._product_info, prompts=self._prompts)
            warehouse_id = installer.configure_warehouse()
            self._config = installer.replace_config(warehouse_id=warehouse_id)
            self._sql_backend = StatementExecutionBackend(self._ws, self._config.warehouse_id)


class AccountInstaller(AccountContext):
    def _get_safe_account_client(self) -> AccountClient:
        """
        Get account client with the correct host based on the cloud provider
        """
        if self.account_client.config.is_account_client:
            return self.account_client
        w = WorkspaceClient(product="ucx", product_version=__version__)
        host = w.config.environment.deployment_url("accounts")
        account_id = self.prompts.question("Please provide the Databricks account id")
        acct_client = AccountClient(host=host, account_id=account_id, product="ucx", product_version=__version__)
        self.replace(account_client=acct_client)
        return self.account_client

    def _get_installer(self, workspace: Workspace) -> WorkspaceInstaller:
        workspace_client = self.account_client.get_workspace_client(workspace)
        return WorkspaceInstaller(workspace_client).replace(product_info=self.product_info, prompts=self.prompts)

    def install_on_account(self):
        ctx = AccountContext(self._get_safe_account_client())
        default_config = None
        confirmed = False
        accessible_workspaces = self.account_workspaces.get_accessible_workspaces()
        msg = "\n".join([w.deployment_name for w in accessible_workspaces])
        installed_workspaces = []
        if not self.prompts.confirm(
            f"UCX has detected the following workspaces available to install. \n{msg}\nDo you want to continue?"
        ):
            return

        for workspace in accessible_workspaces:
            logger.info(f"Installing UCX on workspace {workspace.deployment_name}")
            installer = self._get_installer(workspace)
            if not confirmed:
                default_config = None
            try:
                default_config = installer.run(default_config)
            except RuntimeWarning:
                logger.info("Skipping workspace...")
            # if user confirms to install on remaining workspaces with the same config, don't prompt again
            installed_workspaces.append(workspace)
            if confirmed:
                continue
            confirmed = self.prompts.confirm(
                "Do you want to install UCX on the remaining workspaces with the same config?"
            )

        installed_workspace_ids = [w.workspace_id for w in installed_workspaces if w.workspace_id is not None]
        for workspace in installed_workspaces:
            installer = self._get_installer(workspace)
            installer.replace_config(installed_workspace_ids=installed_workspace_ids)

        # upload the json dump of workspace info in the .ucx folder
        ctx.account_workspaces.sync_workspace_info(installed_workspaces)

    def get_workspace_contexts(
        self, ws: WorkspaceClient, run_as_collection: bool, **named_parameters
    ) -> list[WorkspaceContext]:

        if not run_as_collection:
            return [WorkspaceContext(ws, named_parameters)]
        other_workspace_id = ws.get_workspace_id()
        workspace_contexts = []
        account_client = self._get_safe_account_client()
        acct_ctx = AccountContext(account_client)
        collection_workspace = account_client.workspaces.get(workspace_id=other_workspace_id)
        if not acct_ctx.account_workspaces.can_administer(collection_workspace):
            logger.error(f"User is not workspace admin of collection workspace {other_workspace_id}")
            return []
        installer = self._get_installer(collection_workspace)
        if installer.config.installed_workspace_ids is None:
            logger.error(f"No collection info found in the workspace {other_workspace_id}")
            return []
        for workspace_id in installer.config.installed_workspace_ids:
            workspace = account_client.workspaces.get(workspace_id)
            if not acct_ctx.account_workspaces.can_administer(workspace):
                logger.error(f"User is not workspace admin of workspace {workspace_id}")
                return []
            workspace_client = account_client.get_workspace_client(workspace)
            ctx = WorkspaceContext(workspace_client, named_parameters)
            workspace_contexts.append(ctx)
        return workspace_contexts

    @staticmethod
    def _validate_collection(workspace_ids: list[int], ids_to_workspace: dict[int, Workspace]) -> bool:
        # validate if the passed list of workspace ids, the user has workspace admin access on
        # if not, then cant join a collection
        for workspace_id in workspace_ids:
            if workspace_id not in ids_to_workspace:
                logger.error(
                    f"User doesnt have admin access on the workspace {workspace_id} in the collection, "
                    f"cant join collection."
                )
                return False
        return True

    def join_collection(self, workspace_ids: list[int], join_on_install: bool = False):
        ids_to_workspace: dict[int, Workspace] = {}
        if join_on_install:
            # if run as part of ucx installation prompt user if the installation needs to join a collection
            prompt_message = "Do you want to join the current installation to an existing collection?"
            if not self.prompts.confirm(prompt_message):
                return None

        account_client = self._get_safe_account_client()
        ctx = AccountContext(account_client)
        try:  # pylint: disable=too-many-try-statements
            # if user is account admin list all the available workspace the user has admin access on.
            # This code is run if joining collection after installation or through cli
            accessible_workspaces = ctx.account_workspaces.get_accessible_workspaces()
            for workspace in accessible_workspaces:
                if workspace.workspace_id is not None:
                    ids_to_workspace[workspace.workspace_id] = workspace
            if join_on_install:
                # if run as part of ucx installation allow user to select from the list to join
                target_workspace = self._get_collection_workspace(accessible_workspaces, account_client)
                assert target_workspace is not None and target_workspace.workspace_id is not None
                workspace_ids.append(target_workspace.workspace_id)
        except PermissionDenied:
            # if the user is not account admin, allow user to enter the workspace_id to join as collection.
            # if no workspace_id is entered, then exit
            logger.error("User doesnt have account admin permission, cant list workspaces")
            return None

        if not AccountInstaller._validate_collection(workspace_ids, ids_to_workspace):
            return None
        self._sync_collection(workspace_ids, ids_to_workspace)
        return None

    def _sync_collection(self, workspace_ids: list[int], ids_to_workspace: dict[int, Workspace]):
        # for all the workspace_ids syncs the config.installed_workspace_ids property in all the
        # collection workspaces
        installed_workspace_ids = []
        for workspace_id in workspace_ids:
            workspace = ids_to_workspace[workspace_id]
            installer = self._get_installer(workspace)
            workspace_collection_ids = installer.config.installed_workspace_ids
            if workspace_collection_ids is not None:
                installed_workspace_ids.extend(workspace_collection_ids)
            else:
                installed_workspace_ids.append(workspace_id)
        installed_workspace_ids = list(set(installed_workspace_ids))
        for workspace_id in installed_workspace_ids:
            installed_workspace = ids_to_workspace[workspace_id]
            installer = self._get_installer(installed_workspace)
            installer.replace_config(installed_workspace_ids=installed_workspace_ids)

    def _get_collection_workspace(
        self, accessible_workspaces: list[Workspace], account_client: AccountClient
    ) -> Workspace | None:
        # iterate through each workspace and select workspace which has existing ucx installation
        # allow user to select from the eligible workspace to join as collection
        installed_workspaces = []
        for workspace in accessible_workspaces:
            workspace_client = account_client.get_workspace_client(workspace)
            workspace_installation = Installation.existing(workspace_client, self.product_info.product_name())
            if len(workspace_installation) > 0:
                installed_workspaces.append(workspace)

        if len(installed_workspaces) == 0:
            logger.warning("No existing installation found , setting up new installation without")
            return None
        workspaces = {
            workspace.deployment_name: workspace
            for workspace in installed_workspaces
            if workspace.deployment_name is not None
        }
        target_workspace = self.prompts.choice_from_dict(
            "Please select a workspace, the current installation of ucx will be grouped as a "
            "collection with the selected workspace",
            workspaces,
        )
        return target_workspace


if __name__ == "__main__":
    with_extra("cmd", "install")
    logger = get_logger(__file__)
    if is_in_debug():
        logging.getLogger('databricks').setLevel(logging.DEBUG)
    env = dict(os.environ.items())
    force_install = env.get("UCX_FORCE_INSTALL")
    account_installer = AccountInstaller(AccountClient(product="ucx", product_version=__version__))
    if force_install == "account":
        account_installer.install_on_account()
    else:
        workspace_installer = WorkspaceInstaller(WorkspaceClient(product="ucx", product_version=__version__))
        workspace_installer.run()

        try:
            current_workspace_id = workspace_installer.workspace_client.get_workspace_id()
            account_installer.join_collection([current_workspace_id], True)
        except Unauthenticated:
            logger.warning(
                "User is not authenticated, "
                "you need account admin and workspace admin on "
                "respective workspaces to enable collection joining. Please run join-collection command "
                "with account admin credentials."
            )



import enum
import functools
import http.server
import json
import logging
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from urllib.parse import parse_qsl

from databricks.labs.blueprint.logger import install_logger
from databricks.sdk.service.workspace import Language

from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex, TableMigrationStatus
from databricks.labs.ucx.source_code.base import (
    Advice,
    Advisory,
    Convention,
    Deprecation,
    Failure,
)
from databricks.labs.ucx.source_code.linters.context import LinterContext

logger = logging.getLogger(__name__)


@dataclass
class Position:
    # LSP uses 0-based line and character positions.
    line: int
    character: int

    def as_dict(self) -> dict:
        return {"line": self.line, "character": self.character}

    @classmethod
    def from_dict(cls, raw: dict) -> 'Position':
        return cls(raw['line'], raw['character'])


@dataclass
class Range:
    start: Position
    end: Position

    @classmethod
    def from_dict(cls, raw: dict) -> 'Range':
        return cls(Position.from_dict(raw['start']), Position.from_dict(raw['end']))

    @classmethod
    def make(cls, start_line: int, start_character: int, end_line: int, end_character: int) -> 'Range':
        return cls(start=Position(start_line, start_character), end=Position(end_line, end_character))

    def as_dict(self) -> dict:
        return {"start": self.start.as_dict(), "end": self.end.as_dict()}

    def fragment(self, code: str) -> str:
        out = []
        splitlines = code.splitlines()
        for line, part in enumerate(splitlines):
            if line == self.start.line and line == self.end.line:
                out.append(part[self.start.character : self.end.character])
            elif line == self.start.line:
                out.append(part[self.start.character :])
            elif line == self.end.line:
                out.append(part[: self.end.character])
            elif self.start.line < line < self.end.line:
                out.append(part)
        return "".join(out)


class Severity(enum.IntEnum):
    ERROR = 1
    WARN = 2
    INFO = 3
    HINT = 4


class DiagnosticTag(enum.IntEnum):
    UNNECESSARY = 1
    DEPRECATED = 2


@dataclass
class Diagnostic:
    # the range at which the message applies.
    range: Range

    # The diagnostic's code, which might appear in the user interface.
    code: str

    # An optional property to describe the error code.
    source: str

    # The diagnostic's message.
    message: str

    # The diagnostic's severity. Can be omitted. If omitted it is up to the
    # client to interpret diagnostics as error, warning, info or hint.
    severity: Severity

    tags: list[DiagnosticTag] | None = None

    @classmethod
    def from_advice(cls, advice: Advice) -> 'Diagnostic':
        severity, tags = cls._severity_and_tags(advice)
        return cls(
            range=Range.make(advice.start_line, advice.start_col, advice.end_line, advice.end_col),
            code=advice.code,
            source="databricks.labs.ucx",
            message=advice.message,
            severity=severity,
            tags=tags,
        )

    @classmethod
    def _severity_and_tags(cls, advice):
        if isinstance(advice, Convention):
            return Severity.HINT, [DiagnosticTag.UNNECESSARY]
        if isinstance(advice, Deprecation):
            return Severity.WARN, [DiagnosticTag.DEPRECATED]
        if isinstance(advice, Advisory):
            return Severity.WARN, []
        if isinstance(advice, Failure):
            return Severity.ERROR, []
        return Severity.INFO, []

    def as_dict(self) -> dict:
        return {
            "range": self.range.as_dict(),
            "code": self.code,
            "source": self.source,
            "message": self.message,
            "severity": self.severity.value if self.severity else Severity.WARN,
            "tags": [t.value for t in self.tags] if self.tags else [],
        }


@dataclass
class TextDocumentIdentifier:
    uri: str

    def as_dict(self) -> dict:
        return {
            "uri": self.uri,
        }


@dataclass
class OptionalVersionedTextDocumentIdentifier:
    uri: str
    version: int | None = None

    def as_dict(self) -> dict:
        return {
            "uri": self.uri,
            "version": self.version,
        }


@dataclass
class TextEdit:
    range: Range
    new_text: str

    def as_dict(self) -> dict:
        return {
            "range": self.range.as_dict(),
            "newText": self.new_text,
        }


@dataclass
class TextDocumentEdit:
    text_document: OptionalVersionedTextDocumentIdentifier
    edits: Sequence[TextEdit]

    def as_dict(self) -> dict:
        return {
            "textDocument": self.text_document.as_dict(),
            "edits": [e.as_dict() for e in self.edits],
        }


@dataclass
class WorkspaceEdit:
    # we also can have CreateFile | RenameFile | DeleteFile, but we won't do it for now.
    document_changes: Sequence[TextDocumentEdit]

    def as_dict(self) -> dict:
        return {
            "documentChanges": [e.as_dict() for e in self.document_changes],
        }


@dataclass
class CodeAction:
    title: str
    edit: WorkspaceEdit
    is_preferred: bool = False

    def as_dict(self) -> dict:
        return {
            "title": self.title,
            "edit": self.edit.as_dict(),
            "isPreferred": self.is_preferred,
        }


@dataclass
class AnalyseResponse:
    diagnostics: list[Diagnostic]

    def as_dict(self):
        return {"diagnostics": [d.as_dict() for d in self.diagnostics]}


@dataclass
class QuickFixResponse:
    code_actions: list[CodeAction]

    def as_dict(self):
        return {"code_actions": [ca.as_dict() for ca in self.code_actions]}


class LspServer:
    def __init__(self, language_support: LinterContext):
        self._languages = language_support
        self._extensions = {".py": Language.PYTHON, ".sql": Language.SQL}

    def _read(self, file_uri: str):
        file = Path(file_uri.removeprefix("file://"))
        if file.suffix not in self._extensions:
            raise KeyError(f"no language for {file.suffix}")
        language = self._extensions[file.suffix]
        with file.open('r', encoding='utf8') as f:
            return f.read(), language

    def lint(self, file_uri: str):
        code, language = self._read(file_uri)
        analyser = self._languages.linter(language)
        diagnostics = [Diagnostic.from_advice(_) for _ in analyser.lint(code)]
        return AnalyseResponse(diagnostics)

    def quickfix(self, file_uri: str, code_range: Range, diagnostic_code: str):
        code, language = self._read(file_uri)
        fixer = self._languages.fixer(language, diagnostic_code)
        if not fixer:
            return QuickFixResponse(code_actions=[])
        fragment = code_range.fragment(code)
        apply = fixer.apply(fragment)
        return QuickFixResponse(
            code_actions=[
                CodeAction(
                    title=f"Replace with: {apply}",
                    edit=WorkspaceEdit(
                        document_changes=[
                            TextDocumentEdit(
                                text_document=OptionalVersionedTextDocumentIdentifier(file_uri),
                                edits=[TextEdit(code_range, apply)],
                            ),
                        ]
                    ),
                    is_preferred=True,
                )
            ]
        )

    def serve(self):
        server_address = ('localhost', 8000)
        handler_class = functools.partial(_RequestHandler, self)
        httpd = http.server.ThreadingHTTPServer(server_address, handler_class)
        httpd.serve_forever()


class _RequestHandler(http.server.BaseHTTPRequestHandler):
    def __init__(self, lsp_server: LspServer, *args):
        self._lsp_server = lsp_server
        super().__init__(*args)

    def log_message(self, fmt: str, *args: Any):  # pylint: disable=arguments-differ
        logger.debug(fmt % args)  # pylint: disable=logging-not-lazy

    def do_POST(self):  # pylint: disable=invalid-name
        if not self.path.startswith('/quickfix'):
            self.send_error(400, 'Wrong input')
            return
        self.send_response(200)
        self.end_headers()

        content_length = int(self.headers['Content-Length'])
        post_data = self.rfile.read(content_length)
        raw = json.loads(post_data.decode('utf-8'))
        logger.debug(f"Received:\n{raw}")

        rng = Range.from_dict(raw['range'])
        response = self._lsp_server.quickfix(raw['file_uri'], rng, raw['code'])

        raw = json.dumps(response.as_dict()).encode('utf-8')
        self.wfile.write(raw)

    def do_GET(self):  # pylint: disable=invalid-name
        if not self.path.startswith('/lint'):
            self.send_error(400, 'Wrong input')
            return
        parts = self.path.split('?')
        if len(parts) != 2:
            self.send_error(400, 'Missing Query')
            return
        _, query_string = parts
        query = dict(parse_qsl(query_string))
        response = self._lsp_server.lint(query['file_uri'])
        if not response:
            self.send_error(404, 'no analyser for file type')
            return

        self.send_response(200)
        self.end_headers()
        raw = json.dumps(response.as_dict()).encode('utf-8')
        self.wfile.write(raw)
        self.wfile.flush()


if __name__ == '__main__':
    install_logger()
    logging.root.setLevel('DEBUG')
    languages = LinterContext(
        TableMigrationIndex(
            [
                TableMigrationStatus(
                    src_schema='old', src_table='things', dst_catalog='brand', dst_schema='new', dst_table='stuff'
                ),
                TableMigrationStatus(
                    src_schema='other',
                    src_table='matters',
                    dst_catalog='some',
                    dst_schema='certain',
                    dst_table='issues',
                ),
            ]
        )
    )
    lsp = LspServer(languages)
    lsp.serve()



from functools import cached_property
from os import environ

from databricks.sdk import AccountClient

from databricks.labs.ucx.account.aggregate import AccountAggregate
from databricks.labs.ucx.account.metastores import AccountMetastores
from databricks.labs.ucx.account.workspaces import AccountWorkspaces
from databricks.labs.ucx.contexts.application import CliContext


class AccountContext(CliContext):
    def __init__(self, ac: AccountClient, named_parameters: dict[str, str] | None = None):
        super().__init__(named_parameters)
        self._ac = ac

    @cached_property
    def account_client(self) -> AccountClient:
        return self._ac

    @cached_property
    def workspace_ids(self) -> list[int]:
        return [int(_.strip()) for _ in self.named_parameters.get("workspace_ids", "").split(",") if _]

    @cached_property
    def account_workspaces(self) -> AccountWorkspaces:
        return AccountWorkspaces(self.account_client, self.workspace_ids)

    @cached_property
    def account_aggregate(self) -> AccountAggregate:
        return AccountAggregate(self.account_workspaces)

    @cached_property
    def is_account_install(self) -> bool:
        return environ.get("UCX_FORCE_INSTALL") == "account"

    @cached_property
    def account_metastores(self) -> AccountMetastores:
        return AccountMetastores(self.account_client)



from __future__ import annotations

from pathlib import Path

from databricks.labs.ucx.source_code.graph import Dependency
from databricks.labs.ucx.source_code.linters.context import LinterContext
from databricks.labs.ucx.source_code.notebooks.cells import RunCell
from databricks.labs.ucx.source_code.notebooks.loaders import NotebookLoader
from databricks.labs.ucx.source_code.notebooks.sources import Notebook
from databricks.labs.ucx.source_code.path_lookup import PathLookup


class NotebookMigrator:
    def __init__(self, languages: LinterContext):
        # TODO: move languages to `apply`
        self._languages = languages

    def revert(self, path: Path):
        backup_path = path.with_suffix(".bak")
        if not backup_path.exists():
            return False
        return path.write_text(backup_path.read_text()) > 0

    def apply(self, path: Path) -> bool:
        if not path.exists():
            return False
        dependency = Dependency(NotebookLoader(), path)
        # TODO: the interface for this method has to be changed
        lookup = PathLookup.from_sys_path(Path.cwd())
        container = dependency.load(lookup)
        assert isinstance(container, Notebook)
        return self._apply(container)

    def _apply(self, notebook: Notebook) -> bool:
        changed = False
        for cell in notebook.cells:
            # %run is not a supported language, so this needs to come first
            if isinstance(cell, RunCell):
                # TODO migration data, see https://github.com/databrickslabs/ucx/issues/1327
                if cell.migrate_notebook_path():
                    changed = True
                continue
            if not self._languages.is_supported(cell.language.language):
                continue
            migrated_code = self._languages.apply_fixes(cell.language.language, cell.original_code)
            if migrated_code != cell.original_code:
                cell.migrated_code = migrated_code
                changed = True
        if changed:
            # TODO https://github.com/databrickslabs/ucx/issues/1327 store 'migrated' status
            notebook.path.replace(notebook.path.with_suffix(".bak"))
            notebook.path.write_text(notebook.to_migrated_code())
        return changed
